"""Sketch analyzer plugin for browser search."""
from __future__ import unicode_literals

import logging
import re

import six

from six.moves import urllib_parse as urlparse

from timesketch.lib.analyzers import interface
from timesketch.lib.analyzers import manager
from timesketch.lib.analyzers import utils
from timesketch.lib import emojis


logger = logging.getLogger('timesketch.analyzers.browser_search')


class BrowserSearchSketchPlugin(interface.BaseSketchAnalyzer):
    """Sketch analyzer for BrowserSearch."""

    NAME = 'browser_search'

    DEPENDENCIES = frozenset(['domain'])

    # A list of fields to include in the view output.
    _FIELDS_TO_INCLUDE = ['domain', 'url', 'search_string']

    # Here we define filters and callback methods for all hits on each filter.
    _URL_FILTERS = frozenset([
        ('Bing', re.compile(r'bing\.com/search'),
         '_extract_search_query_from_url', 'q'),
        ('DuckDuckGo', re.compile(r'duckduckgo\.com'),
         '_extract_search_query_from_url', 'q'),
        ('GMail', re.compile(r'mail\.google\.com'),
         '_extract_urlpart_search_query', None),
        ('Google Inbox', re.compile(r'inbox\.google\.com'),
         '_extract_urlpart_search_query', None),
        ('Google Docs', re.compile(r'docs\.google\.com'),
         '_extract_search_query_from_url', 'q'),
        ('Google Groups', re.compile(r'groups\.google\.com/a'),
         '_extract_urlpart_search_query', None),
        ('Google Drive', re.compile(r'drive\.google\.com/.+/search'),
         '_extract_search_query_from_url', 'q'),
        ('Google',
         re.compile(r'(www\.|[a-zA-Z]\.|/)google\.[a-zA-Z]+/search'),
         '_extract_search_query_from_url', 'q'),
        ('Google Sites', re.compile(r'sites\.google\.'),
         '_extract_search_query_from_url', 'q'),
        ('Yahoo', re.compile(r'yahoo\.com/search'),
         '_extract_search_query_from_url', 'p'),
        ('Yandex', re.compile(r'yandex\.com/search'),
         '_extract_search_query_from_url', 'text'),
        ('Youtube', re.compile(r'youtube\.com'),
         '_extract_search_query_from_url', 'search_query'),
    ])

    def __init__(self, index_name, sketch_id):
        """Initialize The Sketch Analyzer.

        Args:
            index_name: Elasticsearch index name
            sketch_id: Sketch ID
        """
        self.index_name = index_name
        super(BrowserSearchSketchPlugin, self).__init__(index_name, sketch_id)

    def _decode_url(self, url):
        """Decodes the URL, replaces %XX to their corresponding characters.

        Args:
          url (str): encoded URL.

        Returns:
          str: decoded URL.
        """
        if not url:
            return ''

        # pylint: disable=too-many-function-args
        decoded_url = urlparse.unquote(url)
        if isinstance(decoded_url, six.binary_type):
            try:
                decoded_url = decoded_url.decode('utf-8')
            except UnicodeDecodeError as exception:
                decoded_url = decoded_url.decode('utf-8', errors='replace')
                logger.warning(
                    'Unable to decode URL: {0:s} with error: {1!s}'.format(
                        url, exception))

        return decoded_url

    def _extract_urlpart_search_query(self, url):
        """Extracts a search query from a URL that uses search/query notation.

        Examples:
            GMail: https://mail.google.com/mail/u/0/#search/query[/?]
            Inbox: https://inbox.google.com/search/<query>
            Groups: https://groups.google.com/a/google.com/forum/#!search/query

        Args:
            url (str): URL.

        Returns:
            str: search query or None if no query was found.
        """
        if 'search/' not in url:
            return None

        _, _, line = url.partition('search/')
        line, _, _ = line.partition('/')
        line, _, _ = line.partition('?')

        search_quoted = line.replace('+', ' ')
        return self._decode_url(search_quoted)

    def _extract_search_query_from_url(self, url, parameter):
        """Extracts a search query from the URL.

        Examples:
            Bing: https://www.bing.com/search?q=query
            GitHub: https://github.com/search?q=query
            Google Drive: https://drive.google.com/drive/search?q=query
            Google Search: https://www.google.com/search?q=query
            Google Sites: https://sites.google.com/site/.*/system/app/pages/
                          search?q=query
            DuckDuckGo: https://duckduckgo.com/?q=query
            Yahoo: https://search.yahoo.com/search?p=query
            Yahoo: https://search.yahoo.com/search;?p=query
            Yandex: https://www.yandex.com/search/?text=query
            YouTube: https://www.youtube.com/results?search_query=query

        Args:
          url (str): URL.
          parameter (str): the parameter that contains the search query.

        Returns:
          str: search query, the search parameter or None if no
              query was found.
        """
        if '{0:s}='.format(parameter) not in url:
            return None

        return self._get_url_parameter_value(url, parameter)

    def _get_url_parameter_value(self, url, parameter):
        """Retrieves the GET parameter from a URL.

        Args:
          url (str): URL.
          parameter (str): the parameter to extract.

        Returns:
          str: the GET parameter value or None if no parameter was found.
        """
        # Make sure we're analyzing the query part of the URL.
        _, _, url = url.partition('?')
        # Look for a key value pair named 'q'.
        _, _, url = url.partition('{0:s}='.format(parameter))
        if not url:
            return ''

        # Strip additional key value pairs.
        parameter, _, _ = url.partition('&')
        parameter = parameter.replace('+', ' ')

        return self._decode_url(parameter)

    def run(self):
        """Entry point for the browser search analyzer.

        Returns:
            String with summary of the analyzer result
        """
        query = 'source_short:"WEBHIST" OR source:"WEBHIST"'
        return_fields = ['url', 'datetime']
        search_emoji = emojis.get_emoji('MAGNIFYING_GLASS')

        # Generator of events based on your query.
        events = self.event_stream(
            query_string=query, return_fields=return_fields)

        simple_counter = 0
        for event in events:
            url = event.source.get('url')

            if url is None:
                continue

            for engine, expression, method_name, parameter in self._URL_FILTERS:
                callback_method = getattr(self, method_name, None)
                if not callback_method:
                    continue

                match = expression.search(url)
                if not match:
                    continue

                if parameter:
                    search_query = callback_method(url, parameter)
                else:
                    search_query = callback_method(url)

                if not search_query:
                    continue

                simple_counter += 1
                datetime = event.source.get('datetime')
                day, _, _ = datetime.partition('T')
                event.add_attributes({
                    'search_string': search_query,
                    'search_engine': engine,
                    'search_day': 'D:{0:s}'.format(day)})

                event.add_human_readable('{0:s} search query: {1:s}'.format(
                    engine, search_query), self.NAME)
                event.add_emojis([search_emoji])
                event.add_tags(['browser-search'])
                # We break at the first hit of a successful search engine.
                break

            # Commit the event to the datastore.
            event.commit()

        if simple_counter > 0:
            view = self.sketch.add_view(
                view_name='Browser Search', analyzer_name=self.NAME,
                query_string='tag:"browser-search"',
                additional_fields=self._FIELDS_TO_INCLUDE)

            params = {
                'field': 'search_string',
                'limit': 20,
                'index': self.index_name,
            }
            agg_obj = self.sketch.add_aggregation(
                name='Top 20 browser search queries ({0:s})'.format(
                    self.timeline_name),
                agg_name='field_bucket', agg_params=params, view_id=view.id,
                chart_type='table',
                description='Created by the browser search analyzer')

            params = {
                'field': 'search_day',
                'index': self.index_name,
                'limit': 20,
            }
            agg_days = self.sketch.add_aggregation(
                name='Top 20 days of search queries ({0:s})'.format(
                    self.timeline_name), agg_name='field_bucket',
                agg_params=params, chart_type='table',
                description='Created by the browser search analyzer',
                label='informational')

            params = {
                'query_string': 'tag:"browser-search"',
                'index': self.index_name,
                'field': 'domain',
            }
            agg_engines = self.sketch.add_aggregation(
                name='Top Search Engines ({0:s})'.format(self.timeline_name),
                agg_name='query_bucket', agg_params=params, view_id=view.id,
                chart_type='hbarchart',
                description='Created by the browser search analyzer')

            story = self.sketch.add_story('{0:s} - {1:s}'.format(
                utils.BROWSER_STORY_TITLE, self.timeline_name))
            story.add_text(utils.BROWSER_STORY_HEADER, skip_if_exists=True)

            story.add_text(
                '## Browser Search Analyzer.\n\nThe browser search '
                'analyzer takes URLs usually resevered for browser '
                'search queries and extracts the search string.'
                'In this timeline the analyzer discovered {0:d} '
                'browser searches.\n\nThis is a summary of '
                'it\'s findings for the timeline **{1:s}**.'.format(
                    simple_counter, self.timeline_name))

            story.add_text(
                'The top 20 most commonly discovered searches were:')
            story.add_aggregation(agg_obj)
            story.add_text('The domains used to search:')
            story.add_aggregation(agg_engines, 'hbarchart')
            story.add_text('And the most common days of search:')
            story.add_aggregation(agg_days)
            story.add_text(
                'And an overview of all the discovered search terms:')
            story.add_view(view)

        return (
            'Browser Search completed with {0:d} search results '
            'extracted.').format(simple_counter)


manager.AnalysisManager.register_analyzer(BrowserSearchSketchPlugin)



"""Sketch analyzer plugin for GCP Service Key usage."""
from __future__ import unicode_literals

from timesketch.lib.analyzers import interface
from timesketch.lib.analyzers import manager


class GcpServiceKeySketchPlugin(interface.BaseSketchAnalyzer):
    """Sketch analyzer for GCP Service Key usage."""

    NAME = 'gcp_servicekey'

    def __init__(self, index_name, sketch_id):
        """Initialize The Sketch Analyzer.

        Args:
            index_name: Elasticsearch index name
            sketch_id: Sketch ID
        """
        self.index_name = index_name
        super(GcpServiceKeySketchPlugin, self).__init__(index_name, sketch_id)

    def run(self):
        """Entry point for the analyzer.

        Returns:
            String with summary of the analyzer result
        """
        # TODO: update dftimewolf stackdriver module to produce more detailed
        # attributes
        query = ('principalEmail:*gserviceaccount.com')
        return_fields = ['message', 'methodName']

        events = self.event_stream(
            query_string=query, return_fields=return_fields)

        simple_counter = 0

        for event in events:
            # Fields to analyze.
            method_name = event.source.get('methodName')

            if 'CreateServiceAccount' in method_name:
                event.add_tags(['service-account-created'])

            if 'compute.instances.insert' in method_name:
                event.add_tags(['gce-instance-created'])

            if 'compute.firewalls.insert' in method_name:
                event.add_tags(['fw-rule-created'])

            if 'compute.networks.insert' in method_name:
                event.add_tags(['network-created'])

            # Commit the event to the datastore.
            event.commit()
            simple_counter += 1

        # Create a saved view with our query.
        if simple_counter:
            self.sketch.add_view(
                view_name='GCP ServiceKey activity', analyzer_name=self.NAME,
                query_string=query)

        return ('GCP ServiceKey analyzer completed with '
                '{0:d} service key marked').format(simple_counter)


manager.AnalysisManager.register_analyzer(GcpServiceKeySketchPlugin)



"""Expert sesssionizer plugins, configured for specific types of activity"""

from __future__ import unicode_literals

from timesketch.lib.analyzers import manager
from timesketch.lib.analyzers.sessionizer import SessionizerSketchPlugin


class WebActivitySessionizerSketchPlugin(SessionizerSketchPlugin):
    """Sessionizer for web activity events"""
    NAME = 'web_activity_sessionizer'
    max_time_diff_micros = 600000000  # 10 minutes
    query = 'source_short:"WEBHIST"'
    session_type = 'web_activity'


class SSHBruteforceSessionizerSketchPlugin(SessionizerSketchPlugin):
    """Sessionizer for SSH bruteforce attacks, where the events that form an
    attack are repeated password failures or attempts to login as a user
    that does not exist."""
    NAME = 'ssh_bruteforce_sessionizer'
    max_time_diff_micros = 10000000 # 10 seconds
    query = 'reporter:"sshd" AND ((message:"invalid user" AND NOT message:' \
        '("keyboard-interactive" OR "connection closed")) OR message:(' \
        '"message repeated" AND "failed password for"))'
    session_type = 'ssh_bruteforce'


manager.AnalysisManager.register_analyzer(WebActivitySessionizerSketchPlugin)
manager.AnalysisManager.register_analyzer(SSHBruteforceSessionizerSketchPlugin)



"""Tests for SigmaPlugin."""
from __future__ import unicode_literals

import mock

from timesketch.lib.analyzers import sigma_tagger
from timesketch.lib.testlib import BaseTest
from timesketch.lib.testlib import MockDataStore


class TestSigmaPlugin(BaseTest):
    """Tests the functionality of the analyzer."""

    def __init__(self, *args, **kwargs):
        super(TestSigmaPlugin, self).__init__(*args, **kwargs)
        self.test_index = 'test_index'


    # Mock the Elasticsearch datastore.
    @mock.patch(
        'timesketch.lib.analyzers.interface.ElasticsearchDataStore',
        MockDataStore)
    def test_analyzer(self):
        """Test analyzer."""
        # TODO: Add more tests

        _ = sigma_tagger.RulesSigmaPlugin(
            sketch_id=1, index_name=self.test_index)



"""Tests for WinEVTXSessionizerSketchPlugin, LogonSessionizerSketchPludin and
UnlockSessionizerSketchPlugin"""

from __future__ import unicode_literals

import unittest
import copy
import mock

from timesketch.lib.analyzers.evtx_sessionizers import \
    LogonSessionizerSketchPlugin
from timesketch.lib.analyzers.evtx_sessionizers import \
    UnlockSessionizerSketchPlugin

from timesketch.lib.analyzers.interface import Event
from timesketch.lib.testlib import BaseTest
from timesketch.lib.testlib import MockDataStore
from timesketch.models.user import User
from timesketch.models.sketch import Sketch

xml_string1 = '<Event xmlns="http://schemas.microsoft.com/win/2004/08/events' \
    '/event"><EventData><Data Name="TargetUserName">USER_1</Data>' \
    '<Data Name="TargetLogonId">0x0000000000000001</Data></EventData></Event>'

xml_string2 = '<Event xmlns="http://schemas.microsoft.com/win/2004/08/events' \
    '/event"><EventData><Data Name="TargetUserName">USER_2</Data>' \
    '<Data Name="TargetLogonId">0x0000000000000002</Data></EventData></Event>'

xml_string3 = '<Event xmlns="http://schemas.microsoft.com/win/2004/08/events' \
    '/event"><EventData><Data Name="TargetUserName">USER_3</Data>' \
    '<Data Name="TargetLogonId">0x0000000000000003</Data></EventData></Event>'

class TestWinEXTXSessionizerPlugin(BaseTest):
    """Tests for Windows EVTX log sessionizers listed in analyzer_classes. New
    sessionizer classes should be added in analyzer_classes, if applicable.

     Attributes:
        analyzer_classes: A list of dictionaries representing analyzer classes
        to test.
    """
    analyzer_classes = [
        {"class": LogonSessionizerSketchPlugin,
         "start_event_id": 4624,
         "end_event_id": 4634},

        {"class": UnlockSessionizerSketchPlugin,
         "start_event_id": 4801,
         "end_event_id": 4800}
    ]

    @mock.patch('timesketch.lib.analyzers.interface.ElasticsearchDataStore',
                MockDataStore)
    def test_get_event_data(self):
        """Test getEventData returns the correct values."""
        user = User('test_user')
        sketch = Sketch('test_sketch', 'description', user)
        label = sketch.Label(label='Test label', user=user)
        sketch.labels.append(label)

        index = 'test_index'
        sketch_id = 1

        for analyzer_class in self.analyzer_classes:
            analyzer = analyzer_class['class'](index, sketch_id)
            datastore = analyzer.datastore
            event_dict = copy.deepcopy(MockDataStore.event_dict)
            event_dict['_source'].update({'xml_string': xml_string1})
            event_obj = Event(event_dict, datastore, sketch)

            username = analyzer.getEventData(event_obj, 'TargetUserName')
            logon_id = analyzer.getEventData(event_obj, 'TargetLogonId')

            self.assertEqual(username, 'USER_1')
            self.assertEqual(logon_id, '0x0000000000000001')

    @mock.patch('timesketch.lib.analyzers.interface.ElasticsearchDataStore',
                MockDataStore)
    def test_one_session(self):
        """Test the behaviour of the analyzer given one start and one end
        event."""
        index = 'test_index'
        sketch_id = 1

        for analyzer_class in self.analyzer_classes:
            analyzer = analyzer_class['class'](index, sketch_id)
            analyzer.datastore.client = mock.Mock()
            datastore = analyzer.datastore

            _create_mock_event(
                datastore,
                0,
                2,
                source_attrs=[{'xml_string': xml_string1,
                               'event_identifier':
                               analyzer_class['start_event_id']},
                              {'xml_string': xml_string1,
                               'event_identifier':
                               analyzer_class['end_event_id']}])

            message = analyzer.run()
            self.assertEqual(
                message,
                'Sessionizing completed, number of sessions created: 1')

            # pylint: disable=unexpected-keyword-arg
            event1 = datastore.event_store['0']
            self.assertEqual(event1['_source']['session_id']
                             [analyzer.session_type], ['0 (USER_1)'])
            event2 = datastore.event_store['1']
            self.assertEqual(event2['_source']['session_id']
                             [analyzer.session_type], ['0 (USER_1)'])

    @mock.patch('timesketch.lib.analyzers.interface.ElasticsearchDataStore',
                MockDataStore)
    def test_multiple_sessions(self):
        """Test multiple sessions are allocated correctly."""
        index = 'test_index'
        sketch_id = 1

        for analyzer_class in self.analyzer_classes:
            analyzer = analyzer_class['class'](index, sketch_id)
            analyzer.datastore.client = mock.Mock()
            datastore = analyzer.datastore

            _create_mock_event(
                datastore,
                0,
                6,
                source_attrs=[{'xml_string': xml_string1,
                               'event_identifier':
                               analyzer_class['start_event_id']},
                              {'xml_string': xml_string1,
                               'event_identifier':
                               analyzer_class['end_event_id']},
                              {'xml_string': xml_string2,
                               'event_identifier':
                               analyzer_class['start_event_id']},
                              {'xml_string': xml_string3,
                               'event_identifier':
                               analyzer_class['start_event_id']},
                              {'xml_string': xml_string2,
                               'event_identifier':
                               analyzer_class['end_event_id']},
                              {'xml_string': xml_string3,
                               'event_identifier':
                               analyzer_class['end_event_id']}])

            message = analyzer.run()
            self.assertEqual(
                message,
                'Sessionizing completed, number of sessions created: 3')

            # pylint: disable=unexpected-keyword-arg

            #session 0
            event1 = datastore.event_store['0']
            self.assertEqual(event1['_source']['session_id']
                             [analyzer.session_type], ['0 (USER_1)'])
            event2 = datastore.event_store['1']
            self.assertEqual(event2['_source']['session_id']
                             [analyzer.session_type], ['0 (USER_1)'])

            #session 1
            event3 = datastore.event_store['2']
            self.assertEqual(event3['_source']['session_id']
                             [analyzer.session_type], ['1 (USER_2)'])
            event5 = datastore.event_store['4']
            self.assertEqual(event5['_source']['session_id']
                             [analyzer.session_type], ['1 (USER_2)'])

            #session 2
            event4 = datastore.event_store['3']
            self.assertEqual(event4['_source']['session_id']
                             [analyzer.session_type], ['2 (USER_3)'])
            event6 = datastore.event_store['5']
            self.assertEqual(event6['_source']['session_id']
                             [analyzer.session_type], ['2 (USER_3)'])

    @mock.patch('timesketch.lib.analyzers.interface.ElasticsearchDataStore',
                MockDataStore)
    def test_startup(self):
        """Test the behaviour of the analyzer with an event stream containing
        a startup event."""
        index = 'test_index'
        sketch_id = 1

        for analyzer_class in self.analyzer_classes:
            analyzer = analyzer_class['class'](index, sketch_id)
            analyzer.datastore.client = mock.Mock()
            datastore = analyzer.datastore

            _create_mock_event(
                datastore,
                0,
                4,
                source_attrs=[{'xml_string': xml_string1,
                               'event_identifier':
                               analyzer_class['start_event_id']},
                              {'xml_string': xml_string2,
                               'event_identifier':
                               analyzer_class['start_event_id']},
                              {'event_identifier': 6005},
                              {'xml_string': xml_string1,
                               'event_identifier':
                               analyzer_class['end_event_id']}])

            message = analyzer.run()
            self.assertEqual(
                message,
                'Sessionizing completed, number of sessions created: 2')

            # pylint: disable=unexpected-keyword-arg
            event1 = datastore.event_store['0']
            self.assertEqual(event1['_source']['session_id']
                             [analyzer.session_type], ['0 (USER_1)'])
            event2 = datastore.event_store['1']
            self.assertEqual(event2['_source']['session_id']
                             [analyzer.session_type], ['1 (USER_2)'])
            event3 = datastore.event_store['2']
            self.assertEqual(set(event3['_source']['session_id']
                                 [analyzer.session_type]),
                             set(['0 (USER_1)', '1 (USER_2)']))
            event4 = datastore.event_store['3']
            self.assertTrue(event4['_source'].get('session_id') is None or
                            event4['_source']['session_id'].get(
                                analyzer.session_type) is None)

    @mock.patch('timesketch.lib.analyzers.interface.ElasticsearchDataStore',
                MockDataStore)
    def test_no_end(self):
        """Test the behaviour of the analyzer given a start event without a
        corresponding end event. The start event is allocated its own session.
        Non-corresponding end events are ignored."""
        index = 'test_index'
        sketch_id = 1

        for analyzer_class in self.analyzer_classes:
            analyzer = analyzer_class['class'](index, sketch_id)
            analyzer.datastore.client = mock.Mock()
            datastore = analyzer.datastore

            _create_mock_event(
                datastore,
                0,
                2,
                source_attrs=[{'xml_string': xml_string1,
                               'event_identifier':
                               analyzer_class['start_event_id']},
                              {'xml_string': xml_string2,
                               'event_identifier':
                               analyzer_class['end_event_id']}])

            message = analyzer.run()
            self.assertEqual(
                message,
                'Sessionizing completed, number of sessions created: 1')

            # pylint: disable=unexpected-keyword-arg
            event1 = datastore.event_store['0']
            self.assertEqual(event1['_source']['session_id']
                             [analyzer.session_type], ['0 (USER_1)'])
            event2 = datastore.event_store['1']
            self.assertTrue(event2['_source'].get('session_id') is None or
                            event2['_source']['session_id'].get(
                                analyzer.session_type) is None)

    @mock.patch('timesketch.lib.analyzers.interface.ElasticsearchDataStore',
                MockDataStore)
    def test_end_before_start(self):
        """Test the behaviour of the analyzer when an end event occurs
        before a start event with the same logon ID. The end event is ignored,
        and the remaining event stream processed normally."""
        index = 'test_index'
        sketch_id = 1

        for analyzer_class in self.analyzer_classes:
            analyzer = analyzer_class['class'](index, sketch_id)
            analyzer.datastore.client = mock.Mock()
            datastore = analyzer.datastore

            _create_mock_event(
                datastore,
                0,
                3,
                source_attrs=[{'xml_string': xml_string1,
                               'event_identifier':
                               analyzer_class['end_event_id']},
                              {'xml_string': xml_string1,
                               'event_identifier':
                               analyzer_class['start_event_id']},
                              {'xml_string': xml_string1,
                               'event_identifier':
                               analyzer_class['end_event_id']}])

            message = analyzer.run()
            self.assertEqual(
                message,
                'Sessionizing completed, number of sessions created: 1')

            # pylint: disable=unexpected-keyword-arg
            event1 = datastore.event_store['0']
            self.assertTrue(event1['_source'].get('session_id') is None or
                            event1['_source']['session_id'].get(
                                analyzer.session_type) is None)
            event2 = datastore.event_store['1']
            self.assertEqual(event2['_source']['session_id']
                             [analyzer.session_type], ['0 (USER_1)'])
            event3 = datastore.event_store['2']
            self.assertEqual(event3['_source']['session_id']
                             [analyzer.session_type], ['0 (USER_1)'])

    @mock.patch('timesketch.lib.analyzers.interface.ElasticsearchDataStore',
                MockDataStore)
    def test_identical_start_events(self):
        """Test the behaviour of the analyzer given multiple start events with
        the same logon ID before a corresponding end event. Each start event is
        allocated its own session, with the end event allocated to the most
        recent start event."""
        index = 'test_index'
        sketch_id = 1

        for analyzer_class in self.analyzer_classes:
            analyzer = analyzer_class['class'](index, sketch_id)
            analyzer.datastore.client = mock.Mock()
            datastore = analyzer.datastore

            _create_mock_event(
                datastore,
                0,
                3,
                source_attrs=[{'xml_string': xml_string1,
                               'event_identifier':
                               analyzer_class['start_event_id']},
                              {'xml_string': xml_string1,
                               'event_identifier':
                               analyzer_class['start_event_id']},
                              {'xml_string': xml_string1,
                               'event_identifier':
                               analyzer_class['end_event_id']}])

            message = analyzer.run()
            self.assertEqual(
                message,
                'Sessionizing completed, number of sessions created: 2')

            # pylint: disable=unexpected-keyword-arg
            event1 = datastore.event_store['0']
            self.assertEqual(event1['_source']['session_id']
                             [analyzer.session_type], ['0 (USER_1)'])
            event2 = datastore.event_store['1']
            self.assertEqual(event2['_source']['session_id']
                             [analyzer.session_type], ['1 (USER_1)'])
            event3 = datastore.event_store['2']
            self.assertEqual(event3['_source']['session_id']
                             [analyzer.session_type], ['1 (USER_1)'])

    @mock.patch('timesketch.lib.analyzers.interface.ElasticsearchDataStore',
                MockDataStore)
    def test_identical_end_events(self):
        """Test the behaviour of the analyzer given multiple end events with
        the same logon ID after a corresponding start event. The first end
        event is allocated to the session, and the other(s) ignored."""
        index = 'test_index'
        sketch_id = 1

        for analyzer_class in self.analyzer_classes:
            analyzer = analyzer_class['class'](index, sketch_id)
            analyzer.datastore.client = mock.Mock()
            datastore = analyzer.datastore

            _create_mock_event(
                datastore,
                0,
                3,
                source_attrs=[{'xml_string': xml_string1,
                               'event_identifier':
                               analyzer_class['start_event_id']},
                              {'xml_string': xml_string1,
                               'event_identifier':
                               analyzer_class['end_event_id']},
                              {'xml_string': xml_string1,
                               'event_identifier':
                               analyzer_class['end_event_id']}])

            message = analyzer.run()
            self.assertEqual(
                message,
                'Sessionizing completed, number of sessions created: 1')

            # pylint: disable=unexpected-keyword-arg
            event1 = datastore.event_store['0']
            self.assertEqual(event1['_source']['session_id']
                             [analyzer.session_type], ['0 (USER_1)'])
            event2 = datastore.event_store['1']
            self.assertEqual(event2['_source']['session_id']
                             [analyzer.session_type], ['0 (USER_1)'])
            event3 = datastore.event_store['2']
            self.assertTrue(event3['_source'].get('session_id') is None or
                            event3['_source']['session_id'].get(
                                analyzer.session_type) is None)

    @mock.patch('timesketch.lib.analyzers.interface.ElasticsearchDataStore',
                MockDataStore)
    def test_zero_events(self):
        """Test the behaviour of the analyzer given an empty event stream."""
        index = 'test_index'
        sketch_id = 1

        for analyzer_class in self.analyzer_classes:
            analyzer = analyzer_class['class'](index, sketch_id)
            analyzer.datastore.client = mock.Mock()
            datastore = analyzer.datastore

            _create_mock_event(datastore, 0, 0)

            message = analyzer.run()
            self.assertEqual(
                message,
                'Sessionizing completed, number of sessions created: 0')

def _create_mock_event(datastore, event_id, quantity, time_diffs=None,
                       source_attrs=None):
    """
    Loads in the datastore mock events that based on the given arguments.

    Args:
        datastore: An instance of MockDataStore.
        event_id: Desired ID for the first Event (to then be incremented).
        quantity: The number of Events to be generated.
        time_diffs: A list of time differences between the generated
        Events.
        source_attrs: A list of attributes to be added to the source attribute
        of the Events.
    """

    if not time_diffs:
        time_diffs = [0]
    quantity = abs(quantity)

    # If the list of time differences is too small to be compatible
    # with the quantity of events, then extend the list with the last
    # value for as many items as necessary.
    if quantity - len(time_diffs) > 0:
        time_diffs.extend([time_diffs[len(time_diffs) - 1]] *
                          (quantity - len(time_diffs)))

    #similarly for source_attrs
    if source_attrs is None:
        source_attrs = [None] * quantity
    else:
        if quantity - len(source_attrs) > 0:
            source_attrs.extend([source_attrs[-1]] *
                                (quantity - len(source_attrs)))

    event_timestamp = 1410895419859714

    for i in range(quantity):
        _create_eventObj(datastore, event_id, event_timestamp, source_attrs[i])
        event_timestamp += abs(time_diffs[i])
        event_id += 1

def _create_eventObj(datastore, event_id, ts, source_attrs):
    """Changes MockDataStore.event_dict based on the given arguments and commits
    it to the datastore.
    Args:
        datastore: An instance of MockDataStore.
        event_id: An event id number.
        ts: A timestamp for an event.
        source_attrs: A dictionary with attributes and theirs values.
    """
    event = copy.deepcopy(MockDataStore.event_dict)
    event['_source']['timestamp'] = ts
    if source_attrs:
        event['_source'].update(source_attrs)

    datastore.import_event(event['_index'], event['_type'], event['_source'],
                           str(event_id))

if __name__ == '__main__':
    unittest.main()



# Copyright 2019 Google Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Barcharts."""

from __future__ import unicode_literals

import altair as alt

from timesketch.lib.charts import manager
from timesketch.lib.charts import interface


class BarChart(interface.BaseChart):
    """Barchart object."""

    NAME = 'barchart'

    def generate(self):
        """Generate the chart.

        Returns:
            Instance of altair.Chart
        """
        chart = self._get_chart_with_transform()
        if self.chart_title:
            chart = chart.mark_bar(strokeWidth=0.3).properties(
                title=self.chart_title)
        else:
            chart = chart.mark_bar(strokeWidth=0.3)

        self._add_url_href(self.encoding)
        chart.encoding = alt.FacetedEncoding.from_dict(self.encoding)
        return chart


class HorizontalBarChart(interface.BaseChart):
    """Horizontal barchart."""

    NAME = 'hbarchart'

    def generate(self):
        """Generate the chart.

        Returns:
            Instance of altair.Chart
        """
        encoding = self.encoding.copy()
        encoding['x'] = self.encoding['y']
        encoding['y'] = self.encoding['x']

        chart = self._get_chart_with_transform()
        self._add_url_href(encoding)

        if self.chart_title:
            bars = chart.mark_bar().properties(
                title=self.chart_title)
        else:
            bars = chart.mark_bar()

        bars.encoding = alt.FacetedEncoding.from_dict(encoding)

        text = bars.mark_text(align='left', baseline='middle', dx=3).encode(
            text='{0:s}:{1:s}'.format(
                encoding['x']['field'], encoding['x']['type']))

        chart = (bars + text)
        return chart


# TODO: Consider making register_chart() accept a list of chart classes.
manager.ChartManager.register_chart(BarChart)
manager.ChartManager.register_chart(HorizontalBarChart)



# Copyright 2017 Google Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for aggregations."""
from __future__ import unicode_literals

import mock

from timesketch.lib.analyzers.similarity_scorer import SimilarityScorer
from timesketch.lib.analyzers.similarity_scorer import SimilarityScorerConfig
from timesketch.lib.testlib import BaseTest
from timesketch.lib.testlib import MockDataStore


class TestSimilarityScorerConfig(BaseTest):
    """Tests for the functionality of the config object."""

    def test_config(self):
        """Test config object."""
        data_type = 'test:test'
        index = 'test_index'
        config = SimilarityScorerConfig(data_type=data_type, index_name=index)

        compare_config = {
            'index_name': '{0}'.format(index),
            'data_type': '{0}'.format(data_type),
            'query': 'data_type:"{0}"'.format(data_type),
            'field': 'message',
            'delimiters': [' ', '-', '/'],
            'threshold': config.DEFAULT_THRESHOLD,
            'num_perm': config.DEFAULT_PERMUTATIONS
        }
        self.assertIsInstance(config, SimilarityScorerConfig)
        for k, v in compare_config.items():
            self.assertEqual(v, getattr(config, k))


class TestSimilarityScorer(BaseTest):
    """Tests for the functionality of the scorer object."""

    def __init__(self, *args, **kwargs):
        super(TestSimilarityScorer, self).__init__(*args, **kwargs)
        self.test_data_type = 'test:test'
        self.test_index = 'test_index'
        self.test_text = 'This is a test text-with tests/test'

    @mock.patch(
        'timesketch.lib.analyzers.interface.ElasticsearchDataStore',
        MockDataStore)
    def test_scorer(self):
        """Test scorer object."""
        scorer = SimilarityScorer(
            index_name=self.test_index,
            sketch_id=1,
            data_type=self.test_data_type)
        self.assertIsInstance(scorer, SimilarityScorer)



# Copyright 2015 Google Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""This module implements the models for the Timesketch core system."""

from __future__ import unicode_literals

import json

from flask import current_app
from flask import url_for

from sqlalchemy import Column
from sqlalchemy import ForeignKey
from sqlalchemy import Integer
from sqlalchemy import Unicode
from sqlalchemy import UnicodeText
from sqlalchemy.orm import relationship

from timesketch.models import BaseModel
from timesketch.models.acl import AccessControlMixin
from timesketch.models.annotations import LabelMixin
from timesketch.models.annotations import CommentMixin
from timesketch.models.annotations import StatusMixin
from timesketch.lib.utils import random_color


class Sketch(AccessControlMixin, LabelMixin, StatusMixin, CommentMixin,
             BaseModel):
    """Implements the Sketch model.

    A Sketch is the collaborative entity in Timesketch. It contains one or more
    timelines that can be grouped and queried on.
    """
    name = Column(Unicode(255))
    description = Column(UnicodeText())
    user_id = Column(Integer, ForeignKey('user.id'))
    timelines = relationship('Timeline', backref='sketch', lazy='select')
    views = relationship('View', backref='sketch', lazy='select')
    events = relationship('Event', backref='sketch', lazy='select')
    stories = relationship('Story', backref='sketch', lazy='select')
    aggregations = relationship('Aggregation', backref='sketch', lazy='select')
    attributes = relationship('Attribute', backref='sketch', lazy='select')
    graphs = relationship('Graph', backref='sketch', lazy='select')
    graphcaches = relationship('GraphCache', backref='sketch', lazy='select')
    aggregationgroups = relationship(
        'AggregationGroup', backref='sketch', lazy='select')
    analysis = relationship('Analysis', backref='sketch', lazy='select')

    def __init__(self, name, description, user):
        """Initialize the Sketch object.

        Args:
            name: The name of the sketch
            description: Description of the sketch
            user: A user (instance of timesketch.models.user.User)
        """
        super(Sketch, self).__init__()
        self.name = name
        self.description = description
        self.user = user

    @property
    def get_named_aggregations(self):
        """Get named aggregations that don't belong to a group.

        Get named aggregations, i.e. only aggregations that have a name and
        are not part of a group.
        """
        return [
            agg for agg in self.aggregations
            if agg.name != '' and not agg.aggregationgroup
        ]

    @property
    def get_named_views(self):
        """Get named views.

        Get named views, i.e. only views that has a name. Views without names
        are used as user state views and should not be visible in the UI.
        """
        views = [
            view for view in self.views
            if view.get_status.status != 'deleted' and view.name != ''
        ]
        return views

    @property
    def external_url(self):
        """Get external URL for the sketch.

        E.g: https://localhost/sketch/42/

        Returns:
            Full URL to the sketch as string.
        """
        url_host = current_app.config.get(
            'EXTERNAL_HOST_URL', 'https://localhost')
        url_path = url_for('sketch_views.overview', sketch_id=self.id)
        return url_host + url_path

    def get_view_urls(self):
        """Get external URL for all views in the sketch.

        Returns:
            Dictionary with url as key and view name as value.
        """

        views = {}
        for view in self.get_named_views:
            url_host = current_app.config.get(
                'EXTERNAL_HOST_URL', 'https://localhost')
            url_path = url_for(
                'sketch_views.explore', sketch_id=self.id, view_id=view.id)
            url = url_host + url_path
            views[url] = view.name
        return views

    @property
    def active_timelines(self):
        """List timelines that are ready for analysis.

        Returns:
            List of instances of timesketch.models.sketch.Timeline
        """
        _timelines = []
        for timeline in self.timelines:
            timeline_status = timeline.get_status.status
            index_status = timeline.searchindex.get_status.status
            if (timeline_status or index_status) in (
                    'processing', 'fail', 'archived'):
                continue
            _timelines.append(timeline)
        return _timelines

    @property
    def get_search_templates(self):
        """Get search templates."""
        return SearchTemplate.query.all()

    def get_user_view(self, user):
        """Get view for user, i.e. view with the state for the user/sketch.

        Args:
            user: User (instance of timesketch.models.user.User)

        Returns:
            view: Instance of timesketch.models.sketch.View
        """
        view = View.query.filter(View.user == user, View.name == '',
                                 View.sketch_id == self.id).order_by(
                                     View.created_at.desc()).first()
        return view


class Timeline(LabelMixin, StatusMixin, CommentMixin, BaseModel):
    """Implements the Timeline model."""
    name = Column(Unicode(255))
    description = Column(UnicodeText())
    color = Column(Unicode(6))
    user_id = Column(Integer, ForeignKey('user.id'))
    searchindex_id = Column(Integer, ForeignKey('searchindex.id'))
    sketch_id = Column(Integer, ForeignKey('sketch.id'))
    analysis = relationship('Analysis', backref='timeline', lazy='select')

    def __init__(self,
                 name,
                 user,
                 sketch,
                 searchindex,
                 color=None,
                 description=None):
        """Initialize the Timeline object.

        Args:
            name: The name of the timeline
            user: A user (instance of timesketch.models.user.User)
            sketch: A sketch (instance of timesketch.models.sketch.Sketch)
            searchindex: A searchindex
                (instance of timesketch.models.sketch.SearchIndex)
            color: Color for the timeline in HEX as string (e.g. F1F1F1F1)
            description: The description for the timeline
        """
        super(Timeline, self).__init__()
        self.name = name
        self.description = description

        if not color:
            color = random_color()

        self.color = color
        self.user = user
        self.sketch = sketch
        self.searchindex = searchindex


class SearchIndex(AccessControlMixin, LabelMixin, StatusMixin, CommentMixin,
                  BaseModel):
    """Implements the SearchIndex model."""
    name = Column(Unicode(255))
    description = Column(UnicodeText())
    index_name = Column(Unicode(255))
    user_id = Column(Integer, ForeignKey('user.id'))
    timelines = relationship(
        'Timeline', backref='searchindex', lazy='dynamic')
    events = relationship('Event', backref='searchindex', lazy='dynamic')

    def __init__(self, name, description, index_name, user):
        """Initialize the SearchIndex object.

        Args:
            name: The name of the timeline
            description: The description for the timeline
            index_name: The name of the searchindex
            user: A user (instance of timesketch.models.user.User)
        """
        super(SearchIndex, self).__init__()
        self.name = name
        self.description = description
        self.index_name = index_name
        self.user = user


class View(AccessControlMixin, LabelMixin, StatusMixin, CommentMixin,
           BaseModel):
    """Implements the View model."""
    name = Column(Unicode(255))
    description = Column(UnicodeText())
    query_string = Column(UnicodeText())
    query_filter = Column(UnicodeText())
    query_dsl = Column(UnicodeText())
    user_id = Column(Integer, ForeignKey('user.id'))
    sketch_id = Column(Integer, ForeignKey('sketch.id'))
    searchtemplate_id = Column(Integer, ForeignKey('searchtemplate.id'))
    aggregations = relationship('Aggregation', backref='view', lazy='select')
    aggregationgroups = relationship(
        'AggregationGroup', backref='view', lazy='select')

    def __init__(self,
                 name,
                 sketch,
                 user,
                 description=None,
                 searchtemplate=None,
                 query_string=None,
                 query_filter=None,
                 query_dsl=None):
        """Initialize the View object.

        Args:
            name: The name of the timeline
            sketch: A sketch (instance of timesketch.models.sketch.Sketch)
            user: A user (instance of timesketch.models.user.User)
            description (str): Description of the view
            searchtemplate: Instance of timesketch.models.sketch.SearchTemplate
            query_string: The query string
            query_filter: The filter to apply (JSON format as string)
            query_dsl: A query DSL document (JSON format as string)
        """
        super(View, self).__init__()
        self.name = name
        self.sketch = sketch
        self.user = user
        self.description = description
        self.searchtemplate = searchtemplate
        self.query_string = query_string
        self.query_filter = query_filter
        self.query_dsl = query_dsl

    def validate_filter(self, query_filter=None):
        """Validate the Query Filter.

        Make sure that we have all expected attributes in the query filter
        json string. The filter dictionary evolves over time and this function
        is used to update all filters.

        Args:
            query_filter: The query filter (JSON format or dictionary)

        Returns:
            query_filter: Query filter dictionary serialized to JSON

        """
        DEFAULT_FROM = 0
        DEFAULT_SIZE = 40 # Number of resulting documents to return
        DEFAULT_LIMIT = DEFAULT_SIZE  # Number of resulting documents to return
        DEFAULT_VALUES = {
            'time_start': None,
            'time_end': None,
            'from': DEFAULT_FROM,
            'size': DEFAULT_SIZE,
            'terminate_after': DEFAULT_LIMIT,
            'indices': [],
            'exclude': [],
            'order': 'asc',
            'chips': []
        }
        # If not provided, get the saved filter from the view
        if not query_filter:
            query_filter = self.query_filter

        # Make sure we have the filter as a dictionary
        if not isinstance(query_filter, dict):
            filter_dict = json.loads(query_filter)
        else:
            filter_dict = query_filter

        # Get all missing attributes and set them to their default value
        missing_attributes = list(
            set(DEFAULT_VALUES.keys()) - set(filter_dict.keys()))
        for key in missing_attributes:
            filter_dict[key] = DEFAULT_VALUES[key]

        return json.dumps(filter_dict, ensure_ascii=False)


class SearchTemplate(AccessControlMixin, LabelMixin, StatusMixin, CommentMixin,
                     BaseModel):
    """Implements the Search Template model."""
    name = Column(Unicode(255))
    description = Column(UnicodeText())
    query_string = Column(UnicodeText())
    query_filter = Column(UnicodeText())
    query_dsl = Column(UnicodeText())
    user_id = Column(Integer, ForeignKey('user.id'))
    views = relationship('View', backref='searchtemplate', lazy='select')

    def __init__(self,
                 name,
                 user,
                 description=None,
                 query_string=None,
                 query_filter=None,
                 query_dsl=None):
        """Initialize the Search Template object.

        Args:
            name: The name of the timeline
            user: A user (instance of timesketch.models.user.User)
            description (str): Description of the search template
            query_string: The query string
            query_filter: The filter to apply (JSON format as string)
            query_dsl: A query DSL document (JSON format as string)
        """
        super(SearchTemplate, self).__init__()
        self.name = name
        self.user = user
        self.description = description
        self.query_string = query_string
        if not query_filter:
            filter_template = {
                'exclude': [],
                'indices': '_all',
                'time_start': None,
                'time_end': None,
                'terminate_after': 40,
                'from': 0,
                'order': 'asc',
                'size': '40'
            }
            query_filter = json.dumps(filter_template, ensure_ascii=False)
        self.query_filter = query_filter
        self.query_dsl = query_dsl


class Event(LabelMixin, StatusMixin, CommentMixin, BaseModel):
    """Implements the Event model."""
    sketch_id = Column(Integer, ForeignKey('sketch.id'))
    searchindex_id = Column(Integer, ForeignKey('searchindex.id'))
    document_id = Column(Unicode(255))

    def __init__(self, sketch, searchindex, document_id):
        """Initialize the Event object.

        Args:
            sketch: A sketch (instance of timesketch.models.sketch.Sketch)
            searchindex: A searchindex
                (instance of timesketch.models.sketch.SearchIndex)
            document_id = String with the datastore document ID
        """
        super(Event, self).__init__()
        self.sketch = sketch
        self.searchindex = searchindex
        self.document_id = document_id


class Story(AccessControlMixin, LabelMixin, StatusMixin, CommentMixin,
            BaseModel):
    """Implements the Story model."""
    title = Column(Unicode(255))
    content = Column(UnicodeText())
    user_id = Column(Integer, ForeignKey('user.id'))
    sketch_id = Column(Integer, ForeignKey('sketch.id'))

    def __init__(self, title, content, sketch, user):
        """Initialize the Story object.

        Args:
            title: The title of the story
            content: Content of the story
            sketch: A sketch (instance of timesketch.models.sketch.Sketch)
            user: A user (instance of timesketch.models.user.User)
        """
        super(Story, self).__init__()
        self.title = title
        self.content = content
        self.sketch = sketch
        self.user = user


class Aggregation(AccessControlMixin, LabelMixin, StatusMixin, CommentMixin,
                  BaseModel):
    """Implements the Aggregation model."""
    name = Column(Unicode(255))
    description = Column(UnicodeText())
    agg_type = Column(Unicode(255))
    parameters = Column(UnicodeText())
    chart_type = Column(Unicode(255))
    user_id = Column(Integer, ForeignKey('user.id'))
    sketch_id = Column(Integer, ForeignKey('sketch.id'))
    view_id = Column(Integer, ForeignKey('view.id'))
    aggregationgroup_id = Column(Integer, ForeignKey('aggregationgroup.id'))

    def __init__(self, name, description, agg_type, parameters, chart_type,
                 user, sketch, view=None, aggregationgroup=None):
        """Initialize the Aggregation object.

        Args:
            name (str): Name of the aggregation
            description (str): Description of the aggregation
            agg_type (str): Aggregation plugin type
            parameters (str): JSON serialized dict with aggregation parameters
            chart_type (str): Chart plugin type
            user (User): The user who created the aggregation
            sketch (Sketch): The sketch that the aggregation is bound to
            view (View): Optional, the view that the aggregation is bound to
            aggregationgroup (AggregationGroup): Optional, an AggregationGroup
                that the aggregation is bound to.
        """
        super(Aggregation, self).__init__()
        self.name = name
        self.description = description
        self.agg_type = agg_type
        self.aggregationgroup = aggregationgroup
        self.parameters = parameters
        self.chart_type = chart_type
        self.user = user
        self.sketch = sketch
        self.view = view


class AggregationGroup(
        AccessControlMixin, LabelMixin, StatusMixin, CommentMixin, BaseModel):
    """Implements the Aggregation Group model."""
    name = Column(Unicode(255))
    description = Column(UnicodeText())
    aggregations = relationship(
        'Aggregation', backref='aggregationgroup', lazy='select')
    parameters = Column(UnicodeText())
    orientation = Column(Unicode(40))
    user_id = Column(Integer, ForeignKey('user.id'))
    sketch_id = Column(Integer, ForeignKey('sketch.id'))
    view_id = Column(Integer, ForeignKey('view.id'))

    def __init__(
            self, name, description, user, sketch, aggregations=None,
            parameters='', orientation='', view=None):
        """Initialize the AggregationGroup object.

        Args:
            name (str): Name of the aggregation
            description (str): Description of the aggregation
            user (User): The user who created the aggregation
            sketch (Sketch): The sketch that the aggregation is bound to
            aggregations (Aggregation): List of aggregation objects.
            parameters (str): A JSON formatted dict with parameters for
                charting.
            orientation (str): Describes how charts should be joined together.
            view (View): Optional: The view that the aggregation is bound to
        """
        super(AggregationGroup, self).__init__()
        self.name = name
        self.description = description
        self.aggregations = aggregations or []
        self.parameters = parameters
        self.orientation = orientation
        self.user = user
        self.sketch = sketch
        self.view = view


class Analysis(LabelMixin, StatusMixin, CommentMixin, BaseModel):
    """Implements the analysis model."""
    name = Column(Unicode(255))
    description = Column(UnicodeText())
    analyzer_name = Column(Unicode(255))
    parameters = Column(UnicodeText())
    result = Column(UnicodeText())
    log = Column(UnicodeText())
    analysissession_id = Column(Integer, ForeignKey('analysissession.id'))
    user_id = Column(Integer, ForeignKey('user.id'))
    sketch_id = Column(Integer, ForeignKey('sketch.id'))
    timeline_id = Column(Integer, ForeignKey('timeline.id'))
    searchindex_id = Column(Integer, ForeignKey('searchindex.id'))

    def __init__(self, name, description, analyzer_name, parameters, user,
                 sketch, timeline=None, searchindex=None, result=None):
        """Initialize the Analysis object.

        Args:
            name (str): Name of the analysis
            description (str): Description of the analysis
            analyzer_name (str): Name of the analyzer
            parameters (str): JSON serialized dict with analyser parameters
            user (User): The user who created the aggregation
            sketch (Sketch): The sketch that the aggregation is bound to
            timeline (Timeline): Timeline the analysis was run on
            searchindex (SearchIndex): SearchIndex the analysis was run on
            result (str): Result report of the analysis
        """
        super(Analysis, self).__init__()
        self.name = name
        self.description = description
        self.analyzer_name = analyzer_name
        self.parameters = parameters
        self.user = user
        self.sketch = sketch
        self.timeline = timeline
        self.searchindex = searchindex
        self.result = result
        self.log = ''


class AnalysisSession(LabelMixin, StatusMixin, CommentMixin, BaseModel):
    """Implements the analysis session model."""
    user_id = Column(Integer, ForeignKey('user.id'))
    sketch_id = Column(Integer, ForeignKey('sketch.id'))
    analyses = relationship(
        'Analysis', backref='analysissession', lazy='select')

    def __init__(self, user, sketch):
        """Initialize the AnalysisSession object.

        Args:
            user (User): The user who created the aggregation
            sketch (Sketch): The sketch that the aggregation is bound to
        """
        super(AnalysisSession, self).__init__()
        self.user = user
        self.sketch = sketch


class Attribute(BaseModel):
    """Implements the attribute model."""
    user_id = Column(Integer, ForeignKey('user.id'))
    sketch_id = Column(Integer, ForeignKey('sketch.id'))
    name = Column(UnicodeText())
    ontology = Column(UnicodeText())
    values = relationship(
        'AttributeValue', backref='attribute', lazy='select')

    def __init__(self, user, sketch, name, ontology):
        """Initialize the Attribute object.

        Args:
            user (User): The user who created the attribute
            sketch (Sketch): The sketch that the attribute is bound to
            name (str): the name of the attribute.
            ontology (str): The ontology of the value, The values that can
                be used are defined in timesketch/lib/ontology.py (ONTOLOGY).
        """
        super(Attribute, self).__init__()
        self.user = user
        self.sketch = sketch
        self.name = name
        self.ontology = ontology


class AttributeValue(BaseModel):
    """Implements the attribute value model."""
    user_id = Column(Integer, ForeignKey('user.id'))
    attribute_id = Column(Integer, ForeignKey('attribute.id'))
    value = Column(UnicodeText())

    def __init__(
            self, user, attribute, value):
        """Initialize the Attribute value object.

        Args:
            user (User): The user who created the attribute value.
            attribute (Attribute): The attribute this value is bound to.
            value (str): a string that contains the value for the attribute.
                The ontology could influence how this will be cast when
                interpreted.
        """
        super(AttributeValue, self).__init__()
        self.user = user
        self.attribute = attribute
        self.value = value


class Graph(LabelMixin, CommentMixin, BaseModel):
    """Implements the graph model."""
    user_id = Column(Integer, ForeignKey('user.id'))
    sketch_id = Column(Integer, ForeignKey('sketch.id'))
    name = Column(UnicodeText())
    description = Column(UnicodeText())
    graph_config = Column(UnicodeText())
    graph_elements = Column(UnicodeText())
    graph_thumbnail = Column(UnicodeText())
    num_nodes = Column(Integer)
    num_edges = Column(Integer)

    def __init__(self, user, sketch, name, description=None, graph_config=None,
                 graph_elements=None, graph_thumbnail=None, num_nodes=None,
                 num_edges=None):
        """Initialize the Graph object.

        Args:
            user (User): The user who created the graph.
            sketch (Sketch): The sketch that the graph is bound to.
            name (str): Name of the graph.
            description (str): Description of the graph.
            graph_config (dict): Config used when generating the graph.
            graph_elements (str): Graph in json string format.
            graph_thumbnail (str): Image of graph in Base64 format.
            num_nodes (int): Number of nodes in the graph.
            num_edges (int): Number of edges in the graph.
        """
        super(Graph, self).__init__()
        self.user = user
        self.sketch = sketch
        self.name = name
        self.description = description
        self.graph_config = graph_config
        self.graph_elements = graph_elements
        self.graph_thumbnail = graph_thumbnail
        self.num_nodes = num_nodes
        self.num_edges = num_edges


class GraphCache(BaseModel):
    """Implements the graph cache model."""
    sketch_id = Column(Integer, ForeignKey('sketch.id'))
    graph_plugin = Column(UnicodeText())
    graph_config = Column(UnicodeText())
    graph_elements = Column(UnicodeText())
    num_nodes = Column(Integer)
    num_edges = Column(Integer)

    def __init__(self, sketch, graph_plugin=None, graph_config=None,
                 graph_elements=None, num_nodes=None, num_edges=None):
        """Initialize the GraphCache object.

        Args:
            sketch (Sketch): The sketch that the graph is bound to.
            graph_plugin (str): Name of the graph plugin that was used.
            graph_config (dict): Config used when generating the graph.
            graph_elements (str): Graph in json string format.
            num_nodes (int): Number of nodes in the graph.
            num_edges (int): Number of edges in the graph.
        """
        super(GraphCache, self).__init__()
        self.sketch = sketch
        self.graph_plugin = graph_plugin
        self.graph_config = graph_config
        self.graph_elements = graph_elements
        self.num_nodes = num_nodes
        self.num_edges = num_edges



"""Sketch analyzer plugin for feature extraction."""
from __future__ import unicode_literals

from timesketch.lib.analyzers import interface
from timesketch.lib.analyzers import manager


class AccountFinderSketchPlugin(interface.BaseSketchAnalyzer):
    """Sketch analyzer for AccountFinder."""

    NAME = 'account_finder'
    DEPENDENCIES = frozenset(['feature_extraction'])

    def __init__(self, index_name, sketch_id):
        """Initialize The Sketch Analyzer.

        Args:
            index_name: Elasticsearch index name
            sketch_id: Sketch ID
        """
        self.index_name = index_name
        super(AccountFinderSketchPlugin, self).__init__(
            index_name, sketch_id)

    def run(self):
        """Entry point for the analyzer.

        Returns:
            String with summary of the analyzer result.
        """
        return_fields = ['found_account', 'tag']

        accounts_found = {}

        events = self.event_stream(
            query_string='_exists_:found_account AND _exists_:tag',
            return_fields=return_fields)

        for event in events:
            event_tags = event.source.get('tag')
            for account_tag in event_tags:
                # There could be other tags on these events; only get the ones
                # related to accounts
                if '-account' not in account_tag:
                    continue

                found_account = event.source.get('found_account')

                accounts_found.setdefault(account_tag, {})
                accounts_found[account_tag].setdefault(found_account, 0)
                accounts_found[account_tag][found_account] += 1

        if accounts_found:
            return (
                '{0:s} identified use of the following accounts: '
                '{1!s}'.format(self.NAME, accounts_found))

        return 'Account finder was unable to extract any accounts.'


manager.AnalysisManager.register_analyzer(AccountFinderSketchPlugin)



# Copyright 2020 Google Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Aggregation resources for version 1 of the Timesketch API."""

import json
import time

from flask import jsonify
from flask import request
from flask import abort
from flask_restful import Resource
from flask_login import login_required
from flask_login import current_user

from timesketch.api.v1 import resources
from timesketch.api.v1.utils import run_aggregator_group
from timesketch.lib import forms
from timesketch.lib.definitions import HTTP_STATUS_CODE_OK
from timesketch.lib.definitions import HTTP_STATUS_CODE_CREATED
from timesketch.lib.definitions import HTTP_STATUS_CODE_BAD_REQUEST
from timesketch.lib.definitions import HTTP_STATUS_CODE_FORBIDDEN
from timesketch.lib.definitions import HTTP_STATUS_CODE_NOT_FOUND
from timesketch.lib.aggregators import manager as aggregator_manager
from timesketch.models import db_session
from timesketch.models.sketch import Aggregation
from timesketch.models.sketch import AggregationGroup
from timesketch.models.sketch import Sketch


class AggregationResource(resources.ResourceMixin, Resource):
    """Resource to query for aggregated results."""

    @login_required
    def get(self, sketch_id, aggregation_id):  # pylint: disable=unused-argument
        """Handles GET request to the resource.

        Handler for /api/v1/sketches/:sketch_id/aggregation/:aggregation_id

        Args:
            sketch_id: Integer primary key for a sketch database model
            aggregation_id: Integer primary key for an agregation database model

        Returns:
            JSON with aggregation results
        """
        sketch = Sketch.query.get_with_acl(sketch_id)
        if not sketch:
            abort(
                HTTP_STATUS_CODE_NOT_FOUND, 'No sketch found with this ID.')
        if not sketch.has_permission(current_user, 'read'):
            abort(HTTP_STATUS_CODE_FORBIDDEN,
                  'User does not have read access controls on sketch.')
        aggregation = Aggregation.query.get(aggregation_id)

        # Check that this aggregation belongs to the sketch
        if aggregation.sketch_id != sketch.id:
            abort(
                HTTP_STATUS_CODE_NOT_FOUND,
                'The sketch ID ({0:d}) does not match with the defined '
                'sketch in the aggregation ({1:d})'.format(
                    aggregation.sketch_id, sketch.id))

        # If this is a user state view, check that it
        # belongs to the current_user
        if aggregation.name == '' and aggregation.user != current_user:
            abort(
                HTTP_STATUS_CODE_FORBIDDEN, (
                    'A user state view can only be viewed by the user it '
                    'belongs to.'))

        return self.to_json(aggregation)

    @login_required
    # pylint: disable=unused-argument
    def post(self, sketch_id, aggregation_id):
        """Handles POST request to the resource.

        Handler for /api/v1/sketches/:sketch_id/aggregation/:aggregation_id

        Args:
            sketch_id: Integer primary key for a sketch database model
            aggregation_id: Integer primary key for an aggregation database
                model
        """
        form = request.json
        if not form:
            form = request.data

        if not form:
            abort(
                HTTP_STATUS_CODE_BAD_REQUEST, 'Unable to validate form data.')

        sketch = Sketch.query.get_with_acl(sketch_id)
        if not sketch:
            abort(
                HTTP_STATUS_CODE_NOT_FOUND, 'No sketch found with this ID.')
        if not sketch.has_permission(current_user, 'write'):
            abort(HTTP_STATUS_CODE_FORBIDDEN,
                  'User does not have write access controls on sketch.')

        aggregation = Aggregation.query.get(aggregation_id)
        if not aggregation:
            abort(
                HTTP_STATUS_CODE_NOT_FOUND,
                'No aggregation found with this ID.')

        if not sketch.has_permission(user=current_user, permission='write'):
            abort(
                HTTP_STATUS_CODE_FORBIDDEN,
                'The user does not have write permission on the sketch.')

        aggregation.name = form.get('name', '')
        aggregation.description = form.get('description', '')
        aggregation.agg_type = form.get('agg_type', aggregation.agg_type)
        aggregation.chart_type = form.get('chart_type', aggregation.chart_type)
        aggregation.user = current_user
        aggregation.sketch = sketch

        labels = form.get('labels', '')
        if labels:
            for label in json.loads(labels):
                if aggregation.has_label(label):
                    continue
                aggregation.add_label(label)

        if form.get('parameters'):
            aggregation.parameters = json.dumps(
                form.get('parameters'), ensure_ascii=False)

        if form.get('view_id'):
            aggregation.view = form.get('view_id')

        db_session.add(aggregation)
        db_session.commit()

        return self.to_json(aggregation, status_code=HTTP_STATUS_CODE_CREATED)


class AggregationInfoResource(resources.ResourceMixin, Resource):
    """Resource to get information about an aggregation class."""

    REMOVE_FIELDS = frozenset(['_shards', 'hits', 'timed_out', 'took'])

    @staticmethod
    def _get_info(aggregator_name):
        """Returns a dict with information about an aggregation."""
        agg_class = aggregator_manager.AggregatorManager.get_aggregator(
            aggregator_name)

        field_lines = []
        for form_field in agg_class.FORM_FIELDS:
            field = {
                'name': form_field.get('name', 'N/A'),
                'description': form_field.get('label', 'N/A')
            }
            field_lines.append(field)

        return {
            'name': agg_class.NAME,
            'display_name': agg_class.DISPLAY_NAME,
            'description': agg_class.DESCRIPTION,
            'fields': field_lines,
        }

    @login_required
    def get(self):
        """Handles GET request to the resource.

        Handler for /api/v1/aggregation/info/

        Returns:
            JSON with information about every aggregator.
        """
        agg_list = []
        for name, _ in aggregator_manager.AggregatorManager.get_aggregators():
            agg_list.append(self._get_info(name))
        return jsonify(agg_list)

    @login_required
    def post(self):
        """Handles POST request to the resource.

        Handler for /api/v1/aggregation/info/

        Returns:
            JSON with aggregation information for a single aggregator.
        """
        form = request.json
        if not form:
            form = request.data

        aggregator_name = form.get('aggregator')
        if not aggregator_name:
            return abort(
                HTTP_STATUS_CODE_BAD_REQUEST,
                'Not able to gather information about an aggregator, '
                'missing the aggregator name.')

        return jsonify(self._get_info(aggregator_name))


class AggregationGroupResource(resources.ResourceMixin, Resource):
    """Resource for aggregation group requests."""

    @login_required
    def get(self, sketch_id, group_id):
        """Handles GET request to the resource.

        Args:
            sketch_id: Integer primary key for a sketch database model.
            group_id: Integer primary key for an aggregation group database
        """
        sketch = Sketch.query.get_with_acl(sketch_id)
        group = AggregationGroup.query.get(group_id)

        if not group:
            abort(
                HTTP_STATUS_CODE_NOT_FOUND, 'No Group found with this ID.')

        if not sketch:
            abort(
                HTTP_STATUS_CODE_NOT_FOUND, 'No sketch found with this ID.')

        if not sketch.has_permission(user=current_user, permission='read'):
            abort(
                HTTP_STATUS_CODE_FORBIDDEN,
                'The user does not have read permission on the sketch.')

        # Check that this group belongs to the sketch
        if group.sketch_id != sketch.id:
            msg = (
                'The sketch ID ({0:d}) does not match with the aggregation '
                'group sketch ID ({1:d})'.format(sketch.id, group.sketch_id))
            abort(HTTP_STATUS_CODE_FORBIDDEN, msg)

        _, objects, meta = run_aggregator_group(group, sketch_id=sketch.id)
        schema = {'meta': meta, 'objects': objects}
        return jsonify(schema)

    @login_required
    def post(self, sketch_id, group_id):
        """Handles POST request to the resource.

        Args:
            sketch_id: Integer primary key for a sketch database model.
            group_id: Integer primary key for an aggregation group database
                model.
        """
        sketch = Sketch.query.get_with_acl(sketch_id)
        group = AggregationGroup.query.get(group_id)
        if not group:
            abort(
                HTTP_STATUS_CODE_NOT_FOUND, 'No Group found with this ID.')

        if not sketch:
            abort(
                HTTP_STATUS_CODE_NOT_FOUND, 'No sketch found with this ID.')

        # Check that this group belongs to the sketch
        if group.sketch_id != sketch.id:
            msg = (
                'The sketch ID ({0:d}) does not match with the aggregation '
                'group sketch ID ({1:d})'.format(sketch.id, group.sketch_id))
            abort(HTTP_STATUS_CODE_FORBIDDEN, msg)

        if not sketch.has_permission(user=current_user, permission='write'):
            abort(
                HTTP_STATUS_CODE_FORBIDDEN,
                'The user does not have write permission on the sketch.')

        form = request.json
        if not form:
            abort(
                HTTP_STATUS_CODE_BAD_REQUEST,
                'No JSON data, unable to process request to create '
                'a new aggregation group.')

        group.name = form.get('name', group.name)
        group.description = form.get('description', group.description)
        group.parameters = form.get('parameters', group.parameters)
        group.orientation = form.get('orientation', group.orientation)
        group.user = current_user
        group.sketch = sketch

        agg_ids = json.loads(form.get('aggregations', group.aggregations))
        aggregations = []

        for agg_id in agg_ids:
            aggregation = Aggregation.query.get(agg_id)
            if not aggregation:
                abort(
                    HTTP_STATUS_CODE_BAD_REQUEST,
                    'No aggregation found for ID: {0:d}'.format(agg_id))
            aggregations.append(aggregation)

        group.aggregations = aggregations

        db_session.add(group)
        db_session.commit()

        return self.to_json(group, status_code=HTTP_STATUS_CODE_CREATED)

    @login_required
    def delete(self, sketch_id, group_id):
        """Handles DELETE request to the resource.

        Args:
            sketch_id: Integer primary key for a sketch database model.
            group_id: Integer primary key for an aggregation group database
                model.
        """
        sketch = Sketch.query.get_with_acl(sketch_id)
        group = AggregationGroup.query.get(group_id)

        if not group:
            abort(
                HTTP_STATUS_CODE_NOT_FOUND, 'No Group found with this ID.')

        if not sketch:
            abort(
                HTTP_STATUS_CODE_NOT_FOUND, 'No sketch found with this ID.')

        # Check that this group belongs to the sketch
        if group.sketch_id != sketch.id:
            msg = (
                'The sketch ID ({0:d}) does not match with the aggregation '
                'group sketch ID ({1:d})'.format(sketch.id, group.sketch_id))
            abort(HTTP_STATUS_CODE_FORBIDDEN, msg)

        if not sketch.has_permission(user=current_user, permission='write'):
            abort(
                HTTP_STATUS_CODE_FORBIDDEN,
                'The user does not have write permission on the sketch.')

        db_session.delete(group)
        db_session.commit()
        return HTTP_STATUS_CODE_OK


class AggregationExploreResource(resources.ResourceMixin, Resource):
    """Resource to send an aggregation request."""

    REMOVE_FIELDS = frozenset(['_shards', 'hits', 'timed_out', 'took'])

    @login_required
    def post(self, sketch_id):
        """Handles POST request to the resource.

        Handler for /api/v1/sketches/<int:sketch_id>/aggregation/explore/

        Args:
            sketch_id: Integer primary key for a sketch database model

        Returns:
            JSON with aggregation results
        """
        form = forms.AggregationExploreForm.build(request)
        if not form.validate_on_submit():
            abort(
                HTTP_STATUS_CODE_BAD_REQUEST,
                'Not able to run aggregation, unable to validate form data.')

        sketch = Sketch.query.get_with_acl(sketch_id)
        if not sketch:
            abort(
                HTTP_STATUS_CODE_NOT_FOUND, 'No sketch found with this ID.')

        if not sketch.has_permission(current_user, 'read'):
            abort(HTTP_STATUS_CODE_FORBIDDEN,
                  'User does not have read access controls on sketch.')

        if sketch.get_status.status == 'archived':
            abort(
                HTTP_STATUS_CODE_BAD_REQUEST,
                'Not able to run aggregation on an archived sketch.')

        sketch_indices = {
            t.searchindex.index_name
            for t in sketch.timelines
            if t.get_status.status.lower() == 'ready'
        }
        indices_string = ','.join(sketch_indices)

        aggregation_dsl = form.aggregation_dsl.data
        aggregator_name = form.aggregator_name.data

        if aggregator_name:
            if isinstance(form.aggregator_parameters.data, dict):
                aggregator_parameters = form.aggregator_parameters.data
            else:
                aggregator_parameters = json.loads(
                    form.aggregator_parameters.data)

            agg_class = aggregator_manager.AggregatorManager.get_aggregator(
                aggregator_name)
            if not agg_class:
                return {}
            if not aggregator_parameters:
                aggregator_parameters = {}

            index = aggregator_parameters.pop('index', indices_string)
            aggregator = agg_class(sketch_id=sketch_id, index=index)
            chart_type = aggregator_parameters.pop('supported_charts', None)
            chart_color = aggregator_parameters.pop('chart_color', '')
            chart_title = aggregator_parameters.pop(
                'chart_title', aggregator.chart_title)
            time_before = time.time()
            result_obj = aggregator.run(**aggregator_parameters)
            time_after = time.time()

            aggregator_description = aggregator.describe

            buckets = result_obj.to_dict()
            buckets['buckets'] = buckets.pop('values')
            result = {
                'aggregation_result': {
                    aggregator_name: buckets
                }
            }
            meta = {
                'method': 'aggregator_run',
                'chart_type': chart_type,
                'name': aggregator_description.get('name'),
                'description': aggregator_description.get('description'),
                'es_time': time_after - time_before,
            }

            if chart_type:
                meta['vega_spec'] = result_obj.to_chart(
                    chart_name=chart_type,
                    chart_title=chart_title, color=chart_color)
                meta['vega_chart_title'] = chart_title

        elif aggregation_dsl:
            # pylint: disable=unexpected-keyword-arg
            result = self.datastore.client.search(
                index=','.join(sketch_indices), body=aggregation_dsl, size=0)

            meta = {
                'es_time': result.get('took', 0),
                'es_total_count': result.get('hits', {}).get('total', 0),
                'timed_out': result.get('timed_out', False),
                'method': 'aggregator_query',
                'max_score': result.get('hits', {}).get('max_score', 0.0)
            }
        else:
            abort(
                HTTP_STATUS_CODE_BAD_REQUEST,
                'An aggregation DSL or a name for an aggregator name needs '
                'to be provided!')

        result_keys = set(result.keys()) - self.REMOVE_FIELDS
        objects = [result[key] for key in result_keys]
        schema = {'meta': meta, 'objects': objects}
        return jsonify(schema)


class AggregationListResource(resources.ResourceMixin, Resource):
    """Resource to query for a list of stored aggregation queries."""

    @login_required
    def get(self, sketch_id):
        """Handles GET request to the resource.

        Handler for /api/v1/sketches/<int:sketch_id>/aggregation/

        Args:
            sketch_id: Integer primary key for a sketch database model

        Returns:
            Views in JSON (instance of flask.wrappers.Response)
        """
        sketch = Sketch.query.get_with_acl(sketch_id)
        if not sketch:
            abort(
                HTTP_STATUS_CODE_NOT_FOUND, 'No sketch found with this ID.')

        if not sketch.has_permission(current_user, 'read'):
            abort(HTTP_STATUS_CODE_FORBIDDEN,
                  'User does not have read access controls on sketch.')
        aggregations = sketch.get_named_aggregations
        return self.to_json(aggregations)

    @staticmethod
    def create_aggregation_from_form(sketch, form):
        """Creates an aggregation from form data.

        Args:
            sketch: Instance of timesketch.models.sketch.Sketch
            form: Instance of timesketch.lib.forms.SaveAggregationForm

        Returns:
            An aggregation (instance of timesketch.models.sketch.Aggregation)
        """
        # Default to user supplied data
        name = form.get('name', '')
        description = form.get('description', '')
        agg_type = form.get('agg_type', '')
        parameter_data = form.get('parameters', {})
        parameters = json.dumps(parameter_data, ensure_ascii=False)
        chart_type = form.get('chart_type', '')
        view_id = form.get('view_id')

        # Create the aggregation in the database
        aggregation = Aggregation(
            name=name,
            description=description,
            agg_type=agg_type,
            parameters=parameters,
            chart_type=chart_type,
            user=current_user,
            sketch=sketch,
            view=view_id
        )

        labels = form.get('labels', '')
        if labels:
            for label in json.loads(labels):
                if aggregation.has_label(label):
                    continue
                aggregation.add_label(label)

        db_session.add(aggregation)
        db_session.commit()

        return aggregation

    @login_required
    def post(self, sketch_id):
        """Handles POST request to the resource.

        Args:
            sketch_id: Integer primary key for a sketch database model

        Returns:
            An aggregation in JSON (instance of flask.wrappers.Response)
        """
        form = request.json
        if not form:
            form = request.data

        if not form:
            abort(
                HTTP_STATUS_CODE_BAD_REQUEST, 'Unable to validate form data.')

        sketch = Sketch.query.get_with_acl(sketch_id)
        if not sketch:
            abort(
                HTTP_STATUS_CODE_NOT_FOUND, 'No sketch found with this ID.')

        if not sketch.has_permission(user=current_user, permission='write'):
            abort(
                HTTP_STATUS_CODE_FORBIDDEN,
                'The user does not have write permission on the sketch.')

        aggregation = self.create_aggregation_from_form(sketch, form)
        return self.to_json(aggregation, status_code=HTTP_STATUS_CODE_CREATED)


class AggregationGroupListResource(resources.ResourceMixin, Resource):
    """Resource to query for a list of stored aggregation queries."""

    @login_required
    def get(self, sketch_id):
        """Handles GET request to the resource.

        Handler for /api/v1/sketches/<int:sketch_id>/aggregation/group/

        Args:
            sketch_id: Integer primary key for a sketch database model

        Returns:
            Views in JSON (instance of flask.wrappers.Response)
        """
        sketch = Sketch.query.get_with_acl(sketch_id)
        if not sketch:
            abort(
                HTTP_STATUS_CODE_NOT_FOUND, 'No sketch found with this ID.')

        if not sketch.has_permission(user=current_user, permission='read'):
            abort(
                HTTP_STATUS_CODE_FORBIDDEN,
                'The user does not have read permission on the sketch.')
        groups = AggregationGroup.query.filter_by(
            sketch_id=sketch_id).all()
        meta = {
            'command': 'list_groups',
        }
        objects = []
        for group in groups:
            group_dict = {
                'id': group.id,
                'name': group.name,
                'parameters': group.parameters,
                'orientation': group.orientation,
                'description': group.description,
                'agg_ids': json.dumps([x.id for x in group.aggregations])
            }
            objects.append(group_dict)
        response = jsonify({'meta': meta, 'objects': objects})
        response.status_code = HTTP_STATUS_CODE_OK
        return response

    @login_required
    def post(self, sketch_id):
        """Handles POST request to the resource.

        Args:
            sketch_id: Integer primary key for a sketch database model

        Returns:
            An aggregation in JSON (instance of flask.wrappers.Response)
        """
        sketch = Sketch.query.get_with_acl(sketch_id)
        if not sketch:
            abort(
                HTTP_STATUS_CODE_NOT_FOUND, 'No sketch found with this ID.')

        if not sketch.has_permission(user=current_user, permission='write'):
            abort(
                HTTP_STATUS_CODE_FORBIDDEN,
                'The user does not have write permission on the sketch.')

        form = request.json
        if not form:
            abort(
                HTTP_STATUS_CODE_BAD_REQUEST,
                'No JSON data, unable to process request to create '
                'a new aggregation group.')

        aggregation_string = form.get('aggregations', '')
        if not aggregation_string:
            abort(
                HTTP_STATUS_CODE_BAD_REQUEST,
                'Unable to create an empty group.')

        agg_list = json.loads(aggregation_string)
        if not isinstance(agg_list, (list, tuple)):
            abort(
                HTTP_STATUS_CODE_BAD_REQUEST,
                'Aggregations needs to be a list of IDs.')

        named_aggs = sketch.get_named_aggregations
        aggregations = [agg for agg in named_aggs if agg.id in agg_list]

        # Create the aggregation in the database
        aggregation_group = AggregationGroup(
            name=form.get('name', 'No Group Name'),
            description=form.get('description', ''),
            parameters=form.get('parameters', ''),
            aggregations=aggregations,
            orientation=form.get('orientation', 'layer'),
            user=current_user,
            sketch=sketch,
            view=form.get('view_id')
        )
        db_session.add(aggregation_group)
        db_session.commit()

        return self.to_json(
            aggregation_group, status_code=HTTP_STATUS_CODE_CREATED)



# Copyright 2017 Google Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Calculate similarity scores based on the Jaccard distance between events."""

from __future__ import unicode_literals

from timesketch.lib import similarity
from timesketch.lib.analyzers import interface
from timesketch.lib.analyzers import manager


class SimilarityScorerConfig(object):
    """Configuration for a similarity scorer."""

    # Parameters for Jaccard and Minhash calculations.
    DEFAULT_THRESHOLD = 0.5
    DEFAULT_PERMUTATIONS = 128

    DEFAULT_CONFIG = {
        'field': 'message',
        'delimiters': [' ', '-', '/'],
        'threshold': DEFAULT_THRESHOLD,
        'num_perm': DEFAULT_PERMUTATIONS
    }

    # For any data_type that need custom config parameters.
    # TODO: Move this to its own file.
    # TODO: Add stopwords boolean config parameter.
    # TODO: Add remove_words boolean config parameter.
    CONFIG_REGISTRY = {
        'windows:evtx:record': {
            'query': 'data_type:"windows:evtx:record"',
            'field': 'message',
            'delimiters': [' ', '-', '/'],
            'threshold': DEFAULT_THRESHOLD,
            'num_perm': DEFAULT_PERMUTATIONS
        }
    }

    def __init__(self, index_name, data_type):
        """Initializes a similarity scorer config.

        Args:
            index_name: Elasticsearch index name.
            data_type: Name of the data_type.
        """
        self._index_name = index_name
        self._data_type = data_type
        for k, v in self._get_config().items():
            setattr(self, k, v)

    def _get_config(self):
        """Get config for supplied data_type.

        Returns:
            Dictionary with configuration parameters.
        """
        config_dict = self.CONFIG_REGISTRY.get(self._data_type)

        # If there is no config for this data_type, use default config and set
        # the query based on the data_type.
        if not config_dict:
            config_dict = self.DEFAULT_CONFIG
            config_dict['query'] = 'data_type:"{0}"'.format(self._data_type)

        config_dict['index_name'] = self._index_name
        config_dict['data_type'] = self._data_type
        return config_dict


class SimilarityScorer(interface.BaseSketchAnalyzer):
    """Score events based on Jaccard distance."""

    NAME = 'similarity_scorer'

    DEPENDENCIES = frozenset()

    def __init__(self, index_name, sketch_id, data_type=None):
        """Initializes a similarity scorer.

        Args:
            index_name: Elasticsearch index name.
            data_type: Name of the data_type.
        """
        self._config = None
        if data_type:
            self._config = SimilarityScorerConfig(index_name, data_type)
        super(SimilarityScorer, self).__init__(index_name, sketch_id)

    def run(self):
        """Entry point for the SimilarityScorer.

        Returns:
            A dict with metadata about the processed data set or None if no
            data_types has been configured.
        """
        if not self._config:
            return 'No data_type specified.'

        # Event generator for streaming results.
        events = self.event_stream(
            query_string=self._config.query,
            return_fields=[self._config.field]
        )

        lsh, minhashes = similarity.new_lsh_index(
            events, field=self._config.field,
            delimiters=self._config.delimiters, num_perm=self._config.num_perm,
            threshold=self._config.threshold)
        total_num_events = len(minhashes)
        for key, minhash in minhashes.items():
            event_id, event_type, index_name = key
            event_dict = dict(_id=event_id, _type=event_type, _index=index_name)
            event = interface.Event(event_dict, self.datastore)
            score = similarity.calculate_score(lsh, minhash, total_num_events)
            attributes_to_add = {'similarity_score': score}
            event.add_attributes(attributes_to_add)
            # Commit the event to the datastore.
            event.commit()

        msg = 'Similarity scorer processed {0:d} events for data_type {1:s}'
        return msg.format(total_num_events, self._config.data_type)


manager.AnalysisManager.register_analyzer(SimilarityScorer)



# Copyright 2015 Google Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for aggregations."""

from __future__ import unicode_literals

from timesketch.lib.aggregators_old import heatmap
from timesketch.lib.testlib import BaseTest
from timesketch.lib.testlib import MockDataStore


class TestAggregators(BaseTest):
    """Tests for the functionality of the aggregation module."""

    def test_heatmap(self):
        """Test to get heatmap data."""
        es_client = MockDataStore('127.0.0,1', 4711)
        h = heatmap(es_client, 1, 'test', {}, [], ['all'])
        self.assertIsInstance(h, list)



#!/usr/bin/env python
"""Main entry point for running tests and linters."""
from __future__ import print_function
from __future__ import unicode_literals

import subprocess
import argparse


def run_python_tests(coverage=False):
    try:
        if coverage:
            subprocess.check_call((
                'nosetests --with-coverage'
                ' --cover-package=timesketch_api_client,'
                'timesketch_import_client,timesketch'
                ' api_client/python/ timesketch/'), shell=True)
        else:
            subprocess.check_call(
                ['nosetests', '-x', 'timesketch/', 'api_client/python/',
                 'importer_client/python'])
    finally:
        subprocess.check_call(['rm', '-f', '.coverage'])


def run_python(args):
    if not args.no_tests:
        run_python_tests(coverage=args.coverage)


def parse_cli_args(args=None):
    """Parse command-line arguments to this script.

    Args:
        args: List of cli arguments not including program name

    Returns:
        Instance of argparse.Namespace with the following boolean attributes:
        py, js, selenium, full, no_tests, coverage

    Raises:
        SystemExit if arguments are invalid or --help is present.
    """
    p = argparse.ArgumentParser(
        description="Run Python unit tests and linters."
    )
    p.add_argument(
        '--no-tests', action='store_true',
        help='Skip tests, run only linters.'
    )
    p.add_argument(
        '--coverage', action='store_true',
        help='Print code coverage report.'
    )
    return p.parse_args(args)


def main():
    args = parse_cli_args()
    run_python(args)


main()



"""Tests for EvtxGapPlugin."""
from __future__ import unicode_literals

import unittest
import mock

from timesketch.lib.analyzers import win_evtxgap
from timesketch.lib.testlib import BaseTest
from timesketch.lib.testlib import MockDataStore

class TestEvtxGapPlugin(BaseTest):
    """Tests the functionality of the analyzer."""

    @mock.patch(
        'timesketch.lib.analyzers.interface.ElasticsearchDataStore',
        MockDataStore)
    def setUp(self):
        super().setUp()
        self.analyzer = win_evtxgap.EvtxGapPlugin('test_index', 1)

    def test_get_range(self):
        """Test the get range function."""
        test_range = [1, 2, 4, 5, 6, 7, 12, 13, 14]
        all_range = list(range(0, 17))

        ranges = list(win_evtxgap.get_range(test_range, all_range))
        expected_ranges = set([(1, 2), (4, 7), (12, 14)])
        self.assertSetEqual(expected_ranges, set(ranges))

        test_range = [0, 3, 4, 5, 6, 7]
        all_range = list(range(0, 17))

        ranges = list(win_evtxgap.get_range(test_range, all_range))
        expected_ranges = set([(0, 0), (3, 7)])
        self.assertSetEqual(expected_ranges, set(ranges))

        test_range = [0, 3, 4, 5, 6, 7]
        all_range = list(range(0, 5))

        with self.assertRaises(IndexError):
            _ = list(win_evtxgap.get_range(test_range, all_range))

        test_range = [0, 3, 4, 5, 6, 7]
        all_range = list(range(2, 50))

        ranges = list(win_evtxgap.get_range(test_range, all_range))
        self.assertSetEqual(set(), set(ranges))


if __name__ == '__main__':
    unittest.main()



# Copyright 2018 Google Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Similarity scorer."""

from __future__ import unicode_literals

import re

from six.moves import filter

from datasketch.minhash import MinHash
from datasketch.lsh import MinHashLSH


# Parameters for Jaccard and Minhash calculations.
DEFAULT_DELIMITERS = [' ', '-', '/']
DEFAULT_THRESHOLD = 0.5
DEFAULT_PERMUTATIONS = 128


def _shingles_from_text(text, delimiters):
    """Splits string into words.

    Args:
        text: string to extract words from.
        delimiters: list of strings used as delimiters for splitting text
            into words.

    Returns:
        List of words.
    """
    # TODO: Remove stopwords using the NLTK python package.
    # TODO: Remove configured patterns from string.
    return list(filter(None, re.split('|'.join(delimiters), text)))


def minhash_from_text(text, num_perm, delimiters):
    """Calculate minhash of text.

    Args:
        text: string to calculate minhash of.
        num_perm: number of random permutation functions used by MinHash to
            be indexed.
        delimiters: list of strings used as delimiters for splitting text
            into words.

    Returns:
        A minhash (instance of datasketch.minhash.MinHash)
    """
    minhash = MinHash(num_perm)
    for word in _shingles_from_text(text, delimiters):
        minhash.update(word.encode('utf8'))
    return minhash


def new_lsh_index(events, field, delimiters=None, num_perm=None,
                  threshold=None):
    """Create a new LSH from a set of Timesketch events.

    Args:
        events: list or an iterator of Event objects.
        field: string denoting the event field to use for the LSH.
        delimiters: list of strings used as delimiters for splitting text
            into words.
        num_perm: number of random permutation functions used by MinHash to
            be indexed.
        threshold: a float for the Jaccard similarity threshold between 0.0 and
            1.0. The initialized MinHash LSH will be optimized for the
            threshold by minizing the false positive and false negative.

    Returns:
        A tuple with an LSH (instance of datasketch.lsh.LSH) and a
        dictionary with event ID as key and minhash as value.
    """
    if delimiters is None:
        delimiters = DEFAULT_DELIMITERS
    if num_perm is None:
        num_perm = DEFAULT_PERMUTATIONS
    if threshold is None:
        threshold = DEFAULT_THRESHOLD

    minhashes = {}
    lsh = MinHashLSH(threshold, num_perm)

    with lsh.insertion_session() as lsh_session:
        for event in events:
            # Insert minhash in LSH index.
            key = (event.event_id, event.event_type, event.index_name)
            minhash = minhash_from_text(
                event.source[field], num_perm, delimiters)
            minhashes[key] = minhash
            lsh_session.insert(key, minhash)

    return lsh, minhashes


def calculate_score(lsh, minhash, total_num_events):
    """Calculate a score based on Jaccard distance.

    The score is calculated based on how many similar events that there are
    for the event being scored. This is called neighbours and we simply
    calculate how many neighbours the event has divided by the total events
    in the LSH.

    Args:
        lsh: instance of datasketch.lsh.MinHashLSH
        minhash: instance of datasketch.minhash.MinHash
        total_num_events: integer of how many events in the LSH

    Returns:
        A float between 0 and 1.
    """
    neighbours = lsh.query(minhash)
    return float(len(neighbours)) / float(total_num_events)



# Copyright 2020 Google Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""SearchIndex resources for version 1 of the Timesketch API."""
import logging

import elasticsearch
from flask import request
from flask import abort
from flask_restful import Resource
from flask_login import login_required
from flask_login import current_user

from timesketch.api.v1 import resources
from timesketch.lib import forms
from timesketch.lib.definitions import HTTP_STATUS_CODE_OK
from timesketch.lib.definitions import HTTP_STATUS_CODE_CREATED
from timesketch.lib.definitions import HTTP_STATUS_CODE_BAD_REQUEST
from timesketch.lib.definitions import HTTP_STATUS_CODE_FORBIDDEN
from timesketch.lib.definitions import HTTP_STATUS_CODE_NOT_FOUND
from timesketch.models import db_session
from timesketch.models.sketch import SearchIndex
from timesketch.models.sketch import Timeline


logger = logging.getLogger('timesketch.index_api')


class SearchIndexListResource(resources.ResourceMixin, Resource):
    """Resource to get all search indices."""

    @login_required
    def get(self):
        """Handles GET request to the resource.

        Returns:
            List of search indices in JSON (instance of flask.wrappers.Response)
        """
        indices = SearchIndex.all_with_acl(current_user).all()
        return self.to_json(
            [i for i in indices if i.get_status.status != 'deleted'])

    @login_required
    def post(self):
        """Handles POST request to the resource.

        Returns:
            A search index in JSON (instance of flask.wrappers.Response)
        """
        form = forms.SearchIndexForm.build(request)
        searchindex_name = form.searchindex_name.data
        es_index_name = form.es_index_name.data
        public = form.public.data

        if not form.validate_on_submit():
            abort(HTTP_STATUS_CODE_BAD_REQUEST, 'Unable to validate form data')

        searchindex = SearchIndex.query.filter_by(
            index_name=es_index_name).first()
        metadata = {'created': True}

        if searchindex:
            metadata['created'] = False
            metadata['deleted'] = searchindex.get_status.status == 'deleted'
            status_code = HTTP_STATUS_CODE_OK
        else:
            searchindex = SearchIndex.get_or_create(
                name=searchindex_name,
                description=searchindex_name,
                user=current_user,
                index_name=es_index_name)
            searchindex.grant_permission(
                permission='read', user=current_user)
            searchindex.grant_permission(
                permission='write', user=current_user)
            searchindex.grant_permission(
                permission='delete', user=current_user)

            if public:
                searchindex.grant_permission(permission='read', user=None)

            # Create the index in Elasticsearch
            self.datastore.create_index(
                index_name=es_index_name, doc_type='generic_event')

            db_session.add(searchindex)
            db_session.commit()
            status_code = HTTP_STATUS_CODE_CREATED

        return self.to_json(
            searchindex, meta=metadata, status_code=status_code)


class SearchIndexResource(resources.ResourceMixin, Resource):
    """Resource to get search index."""

    @login_required
    def get(self, searchindex_id):
        """Handles GET request to the resource.

        Returns:
            Search index in JSON (instance of flask.wrappers.Response)
        """
        searchindex = SearchIndex.query.get_with_acl(searchindex_id)
        return self.to_json(searchindex)

    @login_required
    def delete(self, searchindex_id):
        """Handles DELETE request to the resource."""
        searchindex = SearchIndex.query.get_with_acl(searchindex_id)
        if not searchindex:
            abort(
                HTTP_STATUS_CODE_NOT_FOUND,
                'No searchindex found with this ID.')

        if not searchindex.has_permission(current_user, 'delete'):
            abort(
                HTTP_STATUS_CODE_FORBIDDEN, (
                    'User does not have sufficient access rights to '
                    'delete the search index.'))

        if searchindex.get_status.status == 'deleted':
            abort(
                HTTP_STATUS_CODE_BAD_REQUEST, 'Search index already deleted.')

        timelines = Timeline.query.filter_by(searchindex=searchindex).all()
        sketches = [
            t.sketch for t in timelines
            if t.sketch and t.sketch.get_status.status != 'deleted'
        ]

        if sketches:
            error_strings = ['WARNING: This timeline is in use by:']
            for sketch in sketches:
                error_strings.append(' * {0:s}'.format(sketch.name))
            abort(
                HTTP_STATUS_CODE_FORBIDDEN,
                '\n'.join(error_strings))

        searchindex.set_status(status='deleted')
        db_session.commit()

        other_indexes = SearchIndex.query.filter_by(
            index_name=searchindex.index_name).all()
        if len(other_indexes) > 1:
            logger.warning(
                'Search index: {0:s} belongs to more than one '
                'db entry.'.format(searchindex.index_name))
            return HTTP_STATUS_CODE_OK

        try:
            self.datastore.client.indices.close(index=searchindex.index_name)
        except elasticsearch.NotFoundError:
            logger.warning(
                'Unable to close index: {0:s}, the index wasn\'t '
                'found.'.format(searchindex.index_name))

        return HTTP_STATUS_CODE_OK



# Copyright 2020 Google Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Interface for graphs."""

import hashlib

from flask import current_app
import networkx as nx

from timesketch.lib.datastores.elastic import ElasticsearchDataStore


GRAPH_TYPES = {
    'Graph': nx.Graph,
    'MultiGraph': nx.MultiGraph,
    'DiGraph': nx.DiGraph,
    'MultiDiGraph': nx.MultiDiGraph
}

MAX_EVENTS_PER_EDGE = 500


class Graph:
    """Graph object with helper methods.

    Attributes:
        nx_instance: Networkx graph object.
    """
    def __init__(self, graph_type):
        """Initialize Graph object.

        Args:
            graph_type: (str) Name of graph type.
        """
        _nx_graph_class = GRAPH_TYPES.get(graph_type)
        self.nx_instance = _nx_graph_class()
        self._nodes = {}
        self._edges = {}

    def add_node(self, label, attributes):
        """Add node to graph.

        Args:
            label: (str) Label for the node.
            attributes: (dict) Attributes to add to node.

        Returns:
              Instance of Node object.
        """
        if not attributes:
            attributes = {}

        node = Node(label, attributes)
        node.set_attribute('id', node.id)
        if node.id not in self._nodes:
            self._nodes[node.id] = node
        return node

    def add_edge(self, source, target, label, event, attributes=None):
        """Add edge to graph.

        Args:
            source: (Node) Node to use as source.
            target: (Node) Node to use as target.
            label: (str) Label for the node.
            event: (dict): Elasticsearch event.
            attributes: (dict) Attributes to add to node.
        """
        if not attributes:
            attributes = {}

        attributes['id'] = ''.join([source.id, target.id, label]).lower()

        edge = Edge(source, target, label, attributes)

        if edge.node_counter < MAX_EVENTS_PER_EDGE:
            index = event.get('_index')
            doc_id = event.get('_id')
            events = edge.attributes.get('events', {})
            doc_ids = events.get(index, [])
            doc_ids.append(doc_id)
            edge.node_counter += 1
            events[index] = doc_ids
            edge.set_attribute('events', events)

        self._edges[edge.id] = edge

    def commit(self):
        """Commit all nodes and edges to the networkx graph object."""
        for node_id, node in self._nodes.items():
            self.nx_instance.add_node(
                node_id, label=node.label, **node.attributes)

        for _, edge in self._edges.items():
            label = edge.label + f' ({edge.node_counter})'
            self.nx_instance.add_edge(
                edge.source.id, edge.target.id, label=label,
                **edge.attributes)

    def to_cytoscape(self):
        """Output graph in Cytoscape JSON format.

        Returns:
            Graph in Cytoscape JSON format.
        """
        cy_json = nx.readwrite.json_graph.cytoscape_data(self.nx_instance)
        return cy_json.get('elements', [])


class BaseGraphElement:
    """Base class for graph elements.

    Attributes:
        label (str): Node/Edge label to show in the UI.
        attributes (dict): Attributed to add to the node/edge.
        id (str): Uniq value generated from the label.
    """
    def __init__(self, label='', attributes=None):
        """Initialize the base element object.

        Args:
            label (str): Node/Edge label to show in the UI.
            attributes (dict): Attributes to add to the node/edge.
        """
        self.label = label
        self.attributes = attributes or {}
        self.id = self._generate_id()

    def _generate_id(self):
        """Generate ID for node/edge.

        Returns:
            MD5 hash (str): MD5 hash of the provided label.
        """
        id_string = self.attributes.get('id', self.label)
        return hashlib.md5(id_string.encode('utf-8')).hexdigest()

    def set_attribute(self, key, value):
        """Add or replace an attribute to the element.

        Args:
            key (str): Attribute key.
            value (str): Attribute value.
        """
        self.attributes[key] = value


class Node(BaseGraphElement):
    """Graph node object."""
    # TODO: Add logic for Nodes when needed.


class Edge(BaseGraphElement):
    """Graph edge object.

    Attributes:
        source (Node): Node to add as source node.
        target (Node): Node to add as target node.
        node_counter (int): Counter for number of nodes referenced for the edge.
    """
    def __init__(self, source, target, label='', attributes=None):
        """Initialize the Edge object.

        Args:
            label (str): Node/Edge label to show in the UI.
            attributes (dict): Attributes to add to the edge.
        """
        self.source = source
        self.target = target
        self.node_counter = 0
        super(Edge, self).__init__(label, attributes)


class BaseGraphPlugin:
    """Base class for a graph.

    Attributes:
        datastore (ElasticsearchDataStore): Elasticsearch datastore object.
        graph (nx.Graph): NetworkX Graph object.
    """
    # Name that the graph will be registered as.
    NAME = 'name'

    # Display name (used in the UI)
    DISPLAY_NAME = 'display_name'

    # Description of the plugin (used in the UI)
    DESCRIPTION = 'description'

    # Type of graph. There are four supported types: Undirected Graph,
    # Undirected Multi Graph, Directed Graph, Directed  Multi Graph.
    # If you have multiple edges between nodes you need to use the multi graphs.
    #
    # See NetworkX documentation for details:
    # https://networkx.org/documentation/stable/reference/classes/index.html
    GRAPH_TYPE = 'MultiDiGraph'

    def __init__(self, sketch=None):
        """Initialize the graph object.

        Args:
            sketch (Sketch): Sketch object.

        Raises:
            KeyError if graph type specified is not supported.
        """
        self.datastore = ElasticsearchDataStore(
            host=current_app.config['ELASTIC_HOST'],
            port=current_app.config['ELASTIC_PORT'])
        if not GRAPH_TYPES.get(self.GRAPH_TYPE):
            raise KeyError(f'Graph type {self.GRAPH_TYPE} is not supported')
        self.graph = Graph(self.GRAPH_TYPE)
        self.sketch = sketch

    def _get_all_sketch_indices(self):
        """List all indices in the Sketch.
        Returns:
            List of index names.
        """
        active_timelines = self.sketch.active_timelines
        indices = [t.searchindex.index_name for t in active_timelines]
        return indices

    # TODO: Refactor this to reuse across analyzers and graphs.
    def event_stream(
            self, query_string=None, query_filter=None, query_dsl=None,
            indices=None, return_fields=None, scroll=True):
        """Search ElasticSearch.

        Args:
            query_string: Query string.
            query_filter: Dictionary containing filters to apply.
            query_dsl: Dictionary containing Elasticsearch DSL query.
            indices: List of indices to query.
            return_fields: List of fields to return.
            scroll: Boolean determining whether we support scrolling searches
                or not. Defaults to True.

        Returns:
            Generator of Event objects.

        Raises:
            ValueError: if neither query_string or query_dsl is provided.
        """
        if not (query_string or query_dsl):
            raise ValueError('Both query_string and query_dsl are missing')

        # Query all sketch indices if none are specified.
        if not indices:
            indices = self._get_all_sketch_indices()

        if not query_filter:
            query_filter = {}

        return_fields = list(set(return_fields))

        event_generator = self.datastore.search_stream(
            query_string=query_string,
            query_filter=query_filter,
            query_dsl=query_dsl,
            indices=indices,
            return_fields=return_fields,
            enable_scroll=scroll,
        )
        return event_generator

    def generate(self):
        """Entry point for the graph."""
        raise NotImplementedError



# Copyright 2015 Google Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""This module contains common test utilities for Timesketch."""

from __future__ import unicode_literals

import codecs
import json
import six

from flask_testing import TestCase

from timesketch.app import create_app
from timesketch.lib.definitions import HTTP_STATUS_CODE_REDIRECT
from timesketch.models import init_db
from timesketch.models import drop_all
from timesketch.models import db_session
from timesketch.models.user import Group
from timesketch.models.user import User
from timesketch.models.sketch import Sketch
from timesketch.models.sketch import Timeline
from timesketch.models.sketch import SearchIndex
from timesketch.models.sketch import SearchTemplate
from timesketch.models.sketch import View
from timesketch.models.sketch import Event
from timesketch.models.sketch import Story


class TestConfig(object):
    """Config for the test environment."""
    DEBUG = True
    SECRET_KEY = 'testing'
    SQLALCHEMY_DATABASE_URI = 'sqlite://'
    WTF_CSRF_ENABLED = False
    ELASTIC_HOST = None
    ELASTIC_PORT = None
    LABELS_TO_PREVENT_DELETION = ['protected', 'magic']
    UPLOAD_ENABLED = False
    GRAPH_BACKEND_ENABLED = False
    AUTO_INDEX_ANALYZERS = []
    AUTO_SKETCH_ANALYZERS = []
    SIMILARITY_DATA_TYPES = []


class MockElasticClient(object):
    """A mock implementation of a ElasticSearch client."""

    def __init__(self):
        """Initialize the client."""
        self.indices = MockElasticIndices()

    def search(self, index, body, size):  # pylint: disable=unused-argument
        """Mock a client search, used for aggregations."""
        meta = {
            'es_time': 23,
            'es_total_count': 5621,
            'timed_out': False,
            'max_score': 0.0
        }

        objects = [{
            'my_aggregation': {'buckets': [
                {'foobar': 1, 'second': 'foobar'},
                {'foobar': 4, 'second': 'more stuff'},
                {'foobar': 532, 'second': 'hvernig hefurdu thad'}]},
            'my_second_aggregation': {'buckets': [
                {'foobar': 54, 'second': 'faranlegt', 'third': 'other text'},
                {'foobar': 42, 'second': 'asnalegt'}]}

        }]
        return {'meta': meta, 'objects': objects}


class MockElasticIndices(object):
    # pylint: disable=unused-argument
    def get_mapping(self, *args, **kwargs):
        """Mock get mapping call."""
        return {}

    def stats(self, *args, **kwargs):
        return {'indices': {}}


class MockDataStore(object):
    """A mock implementation of a Datastore."""

    event_dict = {
        '_index': [],
        '_id': 'adc123',
        '_type': 'plaso_event',
        '_source': {
            'es_index': '',
            'es_id': '',
            'label': '',
            'timestamp': 1410895419859714,
            'timestamp_desc': '',
            'datetime': '2014-09-16T19:23:40+00:00',
            'source_short': '',
            'source_long': '',
            'message': '',
        }
    }
    search_result_dict = {
        'hits': {
            'hits': [{
                'sort': [1410593223000],
                '_type': 'plaso_event',
                '_source': {
                    'timestamp':
                    1410593222543942,
                    'message':
                    'Test event',
                    'timesketch_label': [
                        {
                            'user_id': 1,
                            'name': '__ts_star',
                            'sketch_id': 1
                        },
                        {
                            'user_id': 2,
                            'name': '__ts_star',
                            'sketch_id': 99
                        },
                    ],
                    'timestamp_desc':
                    'Content Modification Time',
                    'datetime':
                    '2014-09-13T07:27:03+00:00'
                },
                '_score': 'null',
                '_index': 'test',
                '_id': 'test'
            }],
            'total':
            1,
            'max_score':
            'null'
        },
        '_shards': {
            'successful': 10,
            'failed': 0,
            'total': 10
        },
        'took': 5,
        'timed_out': False
    }

    def __init__(self, host, port):
        """Initialize the datastore.
        Args:
            host: Hostname or IP address to the datastore
            port: The port used by the datastore
        """
        self.client = MockElasticClient()
        self.host = host
        self.port = port
        # Dictionary containing event dictionaries.
        self.event_store = {}

    # pylint: disable=arguments-differ,unused-argument
    def search(self, *args, **kwargs):
        """Mock a search query.
        Returns:
            A dictionary with search result or integer if count is requested.
        """
        if kwargs.get('count'):
            # 4711 is sometimes used instead of 17, on occasions when you want
            # to denote a slightly larger number. Probably comes from the name
            # of 'genuine' Eau-de-cologne, 'No. 4711 ".
            # Ref: https://hack.org/mc/writings/hackerswe/hackerswe.html
            return 4711
        return self.search_result_dict

    def get_event(self, searchindex_id, event_id):
        """Mock returning a single event from the datastore.

        Args:
            searchindex_id: String of ElasticSearch index id
            event_id: String of ElasticSearch event id

        Returns:
            A dictionary with event data.
        """
        return self.event_dict

    @staticmethod
    def get_filter_labels(sketch_id, indices):
        """Mock returning a single event from the datastore.

        Returns:
            A list with label.
        """
        return []

    def set_label(self,
                  searchindex_id,
                  event_id,
                  event_type,
                  sketch_id,
                  user_id,
                  label,
                  toggle=False,
                  single_update=True):
        """Mock adding a label to an event."""
        return

    # pylint: disable=unused-argument
    def create_index(self, *args, **kwargs):
        """Mock creating an index."""
        return

    def import_event(self, index_name, event_type, event=None,
                     event_id=None, flush_interval=None):
        """Mock adding the event to Elasticsearch, instead add the event
        to event_store.
        Args:
            flush_interval: Number of events to queue up before indexing. (This
            functionality is not supported.)
            index_name: Name of the index in Elasticsearch
            event_type: Type of event (e.g. plaso_event)
            event: Event dictionary
            event_id: Event Elasticsearch ID
        """

        if event_id in self.event_store:
            self.event_store[event_id]['_source'].update(event)
            return

        new_event = {
            '_index': index_name,
            '_id': event_id,
            '_type': event_type,
            '_source': event
        }
        self.event_store[event_id] = new_event

    @property
    def version(self):
        """Get Elasticsearch version.

        Returns:
          Version number as a string.
        """
        return '6.0'

    # pylint: disable=unused-argument
    def search_stream(self, query_string, query_filter, query_dsl,
                      indices, return_fields, enable_scroll=True):
        for i in range(len(self.event_store)):
            yield self.event_store[str(i)]


class MockGraphDatabase(object):
    """A mock implementation of a Datastore."""

    def __init__(self, host, username, password):
        """Initialize the datastore.
        Args:
            host: Neo4j host
            username: Neo4j username
            password: Neo4j password
        """
        self.host = host
        self.username = username
        self.password = password

    class MockQuerySequence(object):
        """A mock implementation of a QuerySequence."""
        MOCK_GRAPH = [{
            'nodes': [{
                'id': '1',
                'labels': ['User'],
                'properties': {
                    'username': 'test',
                    'uid': '123456'
                }
            }, {
                'id': '2',
                'labels': ['Machine'],
                'properties': {
                    'hostname': 'test'
                }
            }],
            'relationships': [{
                'endNode': '2',
                'id': '3',
                'startNode': '1',
                'properties': {
                    'method': 'Network'
                },
                'type': 'ACCESS'
            }]
        }]
        MOCK_ROWS = {}
        MOCK_STATS = {}

        def __init__(self):
            self.graph = self.MOCK_GRAPH
            self.rows = self.MOCK_ROWS
            self.stats = self.MOCK_ROWS

    class MockEmptyQuerySequence(object):
        def __init__(self):
            self.graph = None
            self.rows = {}
            self.stats = {}

    # pylint: disable=unused-argument
    def query(self, *args, **kwargs):
        """Mock a search query.
        Returns:
            A MockQuerySequence instance.
        """
        query = args[0]
        if query == 'empty':
            return self.MockEmptyQuerySequence()
        return self.MockQuerySequence()


class BaseTest(TestCase):
    """Base class for tests."""

    COLOR_WHITE = 'FFFFFF'

    def create_app(self):
        """Setup the Flask application.
        Returns:
            Flask application (instance of flask.app.Flask)
        """
        app = create_app(TestConfig)
        return app

    def _commit_to_database(self, model):
        """Add object to the database session and commit.
        Args:
            model: Instance of timesketch.models.[model] object
        """
        db_session.add(model)
        db_session.commit()

    def _create_user(self, username, set_password=False):
        """Create a user in the database.
        Args:
            username: Username (string)
            set_password: Boolean value to decide if a password should be set
        Returns:
            A user (instance of timesketch.models.user.User)
        """
        user = User.get_or_create(username=username)
        if set_password:
            user.set_password(plaintext='test', rounds=4)
        self._commit_to_database(user)
        return user

    def _create_group(self, name, user):
        """Create a user in the database.
        Args:
            name: Group name
            user: A user (instance of timesketch.models.user.User)
        Returns:
            A group (instance of timesketch.models.user.Group)
        """
        group = Group.get_or_create(name=name)
        user.groups.append(group)
        self._commit_to_database(group)
        return group

    def _create_sketch(self, name, user, acl=False):
        """Create a sketch in the database.
        Args:
            name: Name of the sketch (string)
            user: A user (instance of timesketch.models.user.User)
            acl: Boolean value to decide if ACL permissions should be set
        Returns:
            A sketch (instance of timesketch.models.sketch.Sketch)
        """
        sketch = Sketch.get_or_create(name=name, description=name, user=user)
        if acl:
            for permission in ['read', 'write', 'delete']:
                sketch.grant_permission(permission=permission, user=user)
        label = sketch.Label(label='Test label', user=user)
        status = sketch.Status(status='Test status', user=user)
        sketch.labels.append(label)
        sketch.status.append(status)
        self._commit_to_database(sketch)
        return sketch

    def _create_searchindex(self, name, user, acl=False):
        """Create a searchindex in the database.
        Args:
            name: Name of the searchindex (string)
            user: A user (instance of timesketch.models.user.User)
            acl: Boolean value to decide if ACL permissions should be set
        Returns:
            A searchindex (instance of timesketch.models.sketch.SearchIndex)
        """
        searchindex = SearchIndex.get_or_create(
            name=name, description=name, index_name=name, user=user)
        if acl:
            for permission in ['read', 'write', 'delete']:
                searchindex.grant_permission(permission=permission, user=user)
        searchindex.set_status(status='ready')
        self._commit_to_database(searchindex)
        return searchindex

    def _create_event(self, sketch, searchindex, user):
        """Create an event in the database.
        Args:
            sketch: A sketch (instance of timesketch.models.sketch.Sketch)
            searchindex:
                A searchindex (instance of timesketch.models.sketch.SearchIndex)
            user: A user (instance of timesketch.models.user.User)
        Returns:
            An event (instance of timesketch.models.sketch.Event)
        """
        event = Event.get_or_create(
            sketch=sketch, searchindex=searchindex, document_id='test')
        comment = event.Comment(comment='test', user=user)
        event.comments.append(comment)
        self._commit_to_database(event)
        return event

    def _create_story(self, sketch, user):
        """Create a story in the database.
        Args:
            sketch: A sketch (instance of timesketch.models.sketch.Sketch)
            user: A user (instance of timesketch.models.user.User)
        Returns:
            A story (instance of timesketch.models.story.Story)
        """
        story = Story.get_or_create(
            title='Test', content='Test', sketch=sketch, user=user)
        self._commit_to_database(story)
        return story

    def _create_timeline(self, name, sketch, searchindex, user):
        """Create a timeline in the database.
        Args:
            name: Name of the timeline (string)
            sketch: A sketch (instance of timesketch.models.sketch.Sketch)
            searchindex:
                A searchindex (instance of timesketch.models.sketch.SearchIndex)
            user: A user (instance of timesketch.models.user.User)
        Returns:
            A timeline (instance of timesketch.models.sketch.Timeline)
        """
        timeline = Timeline(
            name=name,
            description=name,
            user=user,
            sketch=sketch,
            searchindex=searchindex,
            color=self.COLOR_WHITE)
        timeline.set_status(status='ready')
        self._commit_to_database(timeline)
        return timeline

    def _create_view(self, name, sketch, user):
        """Create a view in the database.
        Args:
            name: Name of the view (string)
            sketch: A sketch (instance of timesketch.models.sketch.Sketch)
            user: A user (instance of timesketch.models.user.User)
        Returns:
            A view (instance of timesketch.models.sketch.View)
        """
        view = View(
            name=name,
            query_string=name,
            query_filter=json.dumps(dict()),
            user=user,
            sketch=sketch)
        self._commit_to_database(view)
        return view

    def _create_searchtemplate(self, name, user):
        """Create a search template in the database.
        Args:
            name: Name of the view (string)
            user: A user (instance of timesketch.models.user.User)
        Returns:
            A search template (timesketch.models.sketch.SearchTemplate)
        """
        searchtemplate = SearchTemplate(
            name=name,
            query_string=name,
            query_filter=json.dumps(dict()),
            user=user)
        self._commit_to_database(searchtemplate)
        return searchtemplate

    def setUp(self):
        """Setup the test database."""
        init_db()

        self.user1 = self._create_user(username='test1', set_password=True)
        self.user2 = self._create_user(username='test2', set_password=False)

        self.group1 = self._create_group(name='test_group1', user=self.user1)
        self.group2 = self._create_group(name='test_group2', user=self.user1)

        self.sketch1 = self._create_sketch(
            name='Test 1', user=self.user1, acl=True)
        self.sketch2 = self._create_sketch(
            name='Test 2', user=self.user1, acl=False)
        self.sketch3 = self._create_sketch(
            name='Test 3', user=self.user1, acl=True)

        self.searchindex = self._create_searchindex(
            name='test', user=self.user1, acl=True)
        self.searchindex2 = self._create_searchindex(
            name='test2', user=self.user1, acl=True)

        self.timeline = self._create_timeline(
            name='Timeline 1',
            sketch=self.sketch1,
            searchindex=self.searchindex,
            user=self.user1)

        self.view1 = self._create_view(
            name='View 1', sketch=self.sketch1, user=self.user1)
        self.view2 = self._create_view(
            name='View 2', sketch=self.sketch2, user=self.user1)
        self.view3 = self._create_view(
            name='', sketch=self.sketch1, user=self.user2)

        self.searchtemplate = self._create_searchtemplate(
            name='template', user=self.user1)

        self.event = self._create_event(
            sketch=self.sketch1, searchindex=self.searchindex, user=self.user1)

        self.story = self._create_story(sketch=self.sketch1, user=self.user1)

    def tearDown(self):
        """Tear down the test database."""
        db_session.remove()
        drop_all()

    def login(self):
        """Authenticate the test user."""
        self.client.post(
            '/login/',
            data=dict(username='test1', password='test'),
            follow_redirects=True)

    def test_unauthenticated(self):
        """
        Generic test for all resources. It tests that no
        unauthenticated request are accepted.
        """
        if not getattr(self, 'resource_url', False):
            self.skipTest(self)

        response = self.client.get(self.resource_url)
        if response.status_code == 405:
            response = self.client.post(self.resource_url)
        if isinstance(response.data, six.binary_type):
            response_data = codecs.decode(response.data, 'utf-8')
        else:
            response_data = response.data
        self.assertIn('/login/', response_data)
        self.assertEqual(response.status_code, HTTP_STATUS_CODE_REDIRECT)


class ModelBaseTest(BaseTest):
    """Base class for database model tests."""

    def _test_db_object(self, expected_result=None, model_cls=None):
        """Generic test that checks if the stored data is correct."""
        db_obj = model_cls.query.get(1)
        for x in expected_result:
            k, v = x[0], x[1]
            self.assertEqual(db_obj.__getattribute__(k), v)



# Copyright 2020 Google Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""End to end tests of Timesketch query functionality."""

from . import interface
from . import manager


class QueryTest(interface.BaseEndToEndTest):
    """End to end tests for query functionality."""

    NAME = 'query_test'

    def __init__(self):
        """Initialize the test."""
        super(QueryTest, self).__init__()

    def setup(self):
        """Import test timeline."""
        self.import_timeline('evtx.plaso')

    def test_wildcard_query(self):
        """Wildcard query over all data in the sketch."""
        response_json = self.sketch.explore(query_string="*")
        count = response_json.get('meta', {}).get('es_total_count', 0)
        self.assertions.assertEqual(count, 3205)


manager.EndToEndTestManager.register_test(QueryTest)



# Copyright 2020 Google Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Graph resources for version 1 of the Timesketch API."""
import logging
import json

from flask_restful import Resource
from flask_login import login_required
from flask_login import current_user

from flask import abort
from flask import jsonify
from flask import request

from timesketch.lib.graphs import manager
from timesketch.api.v1 import resources
from timesketch.models import db_session
from timesketch.models.sketch import Sketch
from timesketch.models.sketch import Graph
from timesketch.models.sketch import GraphCache

from timesketch.lib.definitions import HTTP_STATUS_CODE_NOT_FOUND
from timesketch.lib.definitions import HTTP_STATUS_CODE_FORBIDDEN
from timesketch.lib.definitions import HTTP_STATUS_CODE_BAD_REQUEST

logger = logging.getLogger('timesketch.graph_api')


class GraphListResource(resources.ResourceMixin, Resource):
    """Resource to get all saved graphs for a sketch."""

    @login_required
    def get(self, sketch_id):
        """Handles GET request to the resource.

        Returns:
            List of graphs in JSON (instance of flask.wrappers.Response)
        """
        sketch = Sketch.query.get_with_acl(sketch_id)
        if not sketch:
            abort(HTTP_STATUS_CODE_NOT_FOUND, 'No sketch found with this ID.')

        graphs = sketch.graphs
        return self.to_json(graphs)

    def post(self, sketch_id):
        """Handles POST request to the resource."""
        sketch = Sketch.query.get_with_acl(sketch_id)
        if not sketch:
            abort(HTTP_STATUS_CODE_NOT_FOUND, 'No sketch found with this ID.')

        if not sketch.has_permission(current_user, 'write'):
            abort(HTTP_STATUS_CODE_FORBIDDEN,
                  'User does not have write access on sketch.')

        form = request.json
        name = form.get('name')
        elements = form.get('elements')

        graph = Graph(
            user=current_user, sketch=sketch, name=str(name),
            graph_elements=json.dumps(elements))
        db_session.add(graph)
        db_session.commit()

        return self.to_json(graph)


class GraphResource(resources.ResourceMixin, Resource):
    """Resource to get all graphs."""

    @login_required
    def get(self, sketch_id, graph_id):
        """Handles GET request to the resource.

        Returns:
            List of graphs in JSON (instance of flask.wrappers.Response)
        """
        sketch = Sketch.query.get_with_acl(sketch_id)
        if not sketch:
            abort(HTTP_STATUS_CODE_NOT_FOUND, 'No sketch found with this ID.')

        graph = Graph.query.get(graph_id)
        if not graph:
            abort(HTTP_STATUS_CODE_NOT_FOUND, 'No graph found with this ID.')

        if not sketch.id == graph.sketch.id:
            abort(
                HTTP_STATUS_CODE_BAD_REQUEST,
                'Graph does not belong to this sketch.')

        response = self.to_json(graph).json
        response['objects'][0]['elements'] = graph.graph_elements

        return jsonify(response)


class GraphPluginListResource(resources.ResourceMixin, Resource):
    """Resource to get all graph plugins."""

    @login_required
    def get(self):
        """Handles GET request to the resource.

        Returns:
            List of graphs in JSON (instance of flask.wrappers.Response)
        """
        graph_plugins = manager.GraphManager.get_graphs()
        graphs = []
        for name, graph_class in graph_plugins:
            graph_plugin = {
                'name': name,
                'display_name': graph_class.DISPLAY_NAME,
                'description': graph_class.DESCRIPTION
            }
            graphs.append(graph_plugin)

        return jsonify(graphs)


class GraphCacheResource(resources.ResourceMixin, Resource):
    """Resource to get a cached graph."""

    @login_required
    def post(self, sketch_id):
        """Handles POST request to the resource.

        Returns:
            Graph in JSON (instance of flask.wrappers.Response)
        """
        sketch = Sketch.query.get_with_acl(sketch_id)
        if not sketch:
            abort(HTTP_STATUS_CODE_NOT_FOUND, 'No sketch found with this ID.')

        form = request.json
        plugin_name = form.get('plugin')
        graph_config = form.get('config')
        refresh = form.get('refresh')

        sketch_indices = [
            timeline.searchindex.index_name
            for timeline in sketch.active_timelines
        ]

        cache = GraphCache.get_or_create(
            sketch=sketch, graph_plugin=plugin_name)

        # Refresh cache if timelines have been added/removed from the sketch.
        if cache.graph_config:
            cache_graph_config = json.loads(cache.graph_config)
            if cache_graph_config:
                cache_graph_config = json.loads(cache.graph_config)
                cache_graph_filter = cache_graph_config.get('filter', {})
                cache_filter_indices = cache_graph_filter.get('indices', [])
                if set(sketch_indices) ^ set(cache_filter_indices):
                    refresh = True

        if cache.graph_elements and not refresh:
            return self.to_json(cache)

        graph_class = manager.GraphManager.get_graph(plugin_name)
        graph = graph_class(sketch=sketch)
        cytoscape_json = graph.generate().to_cytoscape()

        if cytoscape_json:
            cache.graph_elements = json.dumps(cytoscape_json)
            cache.graph_config = json.dumps(graph_config)
            cache.update_modification_time()
            db_session.add(cache)
            db_session.commit()

        return self.to_json(cache)



"""Sketch analyzer plugin for Windows crash artefacts."""
from __future__ import unicode_literals

import re

from timesketch.lib.analyzers import interface
from timesketch.lib.analyzers import manager


class WinCrashSketchPlugin(interface.BaseSketchAnalyzer):
    """Sketch analyzer for Windows application crashes."""

    NAME = 'win_crash'

    DEPENDENCIES = frozenset()

    FILENAME_REGEX = re.compile(
        r'(?:\\|\/)(?:AppCrash|Critical|NonCritical)_(.+?\.exe)_[a-f0-9]{16,}'
        r'|\'([^\']+\.exe)\'', re.IGNORECASE)

    QUERY_ELEMENTS = {
        'Event - App Error' : (
            'data_type:"windows:evtx:record"',
            'source_name:"Application Error"',
            'event_identifier:"1000"',
            'event_level:"2"', # Level: Error
        ),
        'Event - WER' : (
            'data_type:"windows:evtx:record"',
            'source_name:"Windows Error Reporting"',
            'event_identifier:"1001"',
            'event_level:"4"', # Level: Info
        ),
        'Event - BSOD' : (
            'data_type:"windows:evtx:record"',
            'source_name:"Microsoft-Windows-WER-SystemErrorReporting"',
            'event_identifier:"1001"',
            'event_level:"2"', # Level: Error
        ),
        'Event - App Hang' : (
            'data_type:"windows:evtx:record"',
            'source_name:"Application Error"',
            'event_identifier:"1002"',
            'event_level:"2"', # Level: Error
        ),
        'Event - EMET Warning' : (
            'data_type:"windows:evtx:record"',
            'source_name:"EMET"',
            'event_identifier:"1"',
            'event_level:"3"', # Level: Warning
        ),
        'Event - EMET Error' : (
            'data_type:"windows:evtx:record"',
            'source_name:"EMET"',
            'event_identifier:"1"',
            'event_level:"2"', # Level: Error
        ),
        'Event - .NET App Crash' : (
            'data_type:"windows:evtx:record"',
            'source_name:".NET Runtime"',
            'event_identifier:"1026"',
            'event_level:"2"', # Level: Error
        ),
        'File - WER Report' : (
            'data_type:"fs:stat"',
            'filename:"/Microsoft/Windows/WER/"',
            'filename:/((Non)?Critical|AppCrash)_.*/',
            'file_entry_type:("directory" or "2")',
        ),
        'Registry - Crash Reporting' : (
            'data_type:"windows:registry:key_value"',
            r'key_path:"\\Control\\CrashControl"',
            'values:("LogEvent: REG_DWORD_LE 0"' + \
            ' OR "SendAlert: REG_DWORD_LE 0"' + \
            ' OR "CrashDumpEnabled: REG_DWORD_LE 0")',
        ),
        'Registry - Error Reporting' : (
            'data_type:"windows:registry:key_value"',
            r'key_path:"\\Software\\Microsoft\\PCHealth\\ErrorReporting"',
            'values:("DoReport: REG_DWORD_LE 0" OR "ShowUI: REG_DWORD_LE 0")',
        ),
    }

    def __init__(self, index_name, sketch_id):
        """Initialize The Sketch Analyzer.

        Args:
            index_name: Elasticsearch index name
            sketch_id: Sketch ID
        """
        self.index_name = index_name
        super(WinCrashSketchPlugin, self).__init__(index_name, sketch_id)

    def formulate_query(self, elements):
        """Generates the Elasticsearch query.

        Args:
            elements: Dictionary with a list of conditions

        Returns:
            The Elasticsearch query
        """
        conditions = list()
        for element_list in elements.values():
            conditions += ['({0})'.format(' AND '.join(element_list))]
        return ' OR '.join(conditions)

    def extract_filename(self, text):
        """Finds filenames of crashed applications using a regular expression.

        Args:
            text: String that might contain the filename

        Returns:
            The string with filename if found
        """
        if '.exe' in str(text or '').lower():
            match = self.FILENAME_REGEX.search(text)
            if match:
                # The regex can match on full file paths and filenames,
                # so only return the filename.
                return min([m for m in match.groups() if m])
        return ''

    def mark_as_crash(self, event, filename):
        """Mark entries with crash artefacts.

        Args:
            event: Elasticsearch event
            filename: Application that crashed
        """
        if filename:
            event.add_attributes({'crash_app': filename})
        event.add_tags(['win_crash'])

    def run(self):
        """Entry point for the analyzer.

        Returns:
            String with summary of the analyzer result
        """
        query = self.formulate_query(self.QUERY_ELEMENTS)

        return_fields = ['data_type', 'message', 'filename']

        # Generator of events based on your query.
        events = self.event_stream(
            query_string=query, return_fields=return_fields)

        # Container for filenames of crashed applications.
        filenames = set()

        for event in events:
            data_type = event.source.get('data_type')
            event_text = None

            # Tag entries that show the crash reporting has been disabled.
            if data_type == 'windows:registry:key_value':
                event.add_comment(
                    'WARNING: The crash reporting was disabled. '
                    'It could be indicative of attacker activity. '
                    'For details refer to page 16 from '
                    'https://assets.documentcloud.org/documents/3461560/'
                    'Google-Aquarium-Clean.pdf.')
                event.commit()
                continue

            # Search event log entries for filenames of crashed applications.
            if data_type == 'windows:evtx:record':
                event_text = event.source.get('message')

            # Search file system entries for filenames of crashed applications.
            elif data_type == 'fs:stat':
                event_text = event.source.get('filename')

            # If found the filename, tag the entry as crash-related
            filename = self.extract_filename(event_text)
            if filename:
                self.mark_as_crash(event, filename)
                filenames.add(filename)
                event.commit()

        # Create a saved view with our query.
        if filenames:
            self.sketch.add_view(
                'Windows Crash activity', 'win_crash',
                query_string='tag:"win_crash"')

        return 'Windows Crash analyzer completed, ' + \
            '{0:d} crashed application{1:s} identified: {2:s}'.format(
                len(filenames),
                's' if len(filenames) > 1 else '',
                ', '.join(filenames))


manager.AnalysisManager.register_analyzer(WinCrashSketchPlugin)



# Copyright 2015 Google Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for the sketch models."""

from __future__ import unicode_literals

import json

from timesketch.models.sketch import Sketch
from timesketch.models.sketch import Timeline
from timesketch.models.sketch import SearchIndex
from timesketch.models.sketch import SearchTemplate
from timesketch.models.sketch import View
from timesketch.models.sketch import Event
from timesketch.models.sketch import Story
from timesketch.lib.testlib import ModelBaseTest


class SketchModelTest(ModelBaseTest):
    """Tests the sketch models."""

    def test_sketch_model(self):
        """
        Test that the test sketch has the expected data stored in the
        database.
        """
        expected_result = frozenset([
            ('name', 'Test 1'),
            ('description', 'Test 1'),
            ('user', self.user1)])
        self._test_db_object(expected_result=expected_result, model_cls=Sketch)

    def test_searchindex_model(self):
        """
        Test that the test searchindex has the expected data stored in the
        database.
        """
        expected_result = frozenset(
            [('name', 'test'), ('description', 'test'),
             ('index_name', 'test'), ('user', self.user1)])
        self._test_db_object(
            expected_result=expected_result, model_cls=SearchIndex)

    def test_timeline_model(self):
        """
        Test that the test timeline has the expected data stored in the
        database.
        """
        expected_result = frozenset([
            ('name', 'Timeline 1'),
            ('description', 'Timeline 1'),
            ('user', self.user1),
            ('sketch', self.sketch1),
            ('searchindex', self.searchindex)])
        self._test_db_object(
            expected_result=expected_result, model_cls=Timeline)

    def test_view_model(self):
        """
        Test that the test view has the expected data stored in the database.
        """
        expected_result = frozenset(
            [('name', 'View 1'), ('query_string', 'View 1'),
             ('user', self.user1), ('sketch', self.sketch1)])
        self._test_db_object(expected_result=expected_result, model_cls=View)

    def test_searchtemplate_model(self):
        """
        Test that the test search template has the expected data stored in the
        database.
        """
        expected_result = frozenset([
            ('name', 'template'),
            ('user', self.user1),
        ])
        self._test_db_object(
            expected_result=expected_result, model_cls=SearchTemplate)

    def test_event_model(self):
        """
        Test that the test event has the expected data stored in the database.
        """
        expected_result = frozenset([('searchindex', self.searchindex),
                                     ('sketch', self.sketch1),
                                     ('document_id', 'test')])
        self._test_db_object(expected_result=expected_result, model_cls=Event)

    def test_validate_filter(self):
        """
         Test the query filter validation.
         """
        DEFAULT_LIMIT = 40
        default_values = {
            'from': 0,
            'time_start': None,
            'time_end': None,
            'size': DEFAULT_LIMIT,
            'terminate_after': DEFAULT_LIMIT,
            'indices': [],
            'exclude': [],
            'order': 'asc',
            'chips': []
        }
        test_filter = dict(indices=[])
        test_filter_json = json.dumps(test_filter)
        validated_filter_dict = self.view1.validate_filter(test_filter)
        validated_filter_json = self.view1.validate_filter(test_filter_json)
        self.assertDictEqual(json.loads(validated_filter_dict), default_values)
        self.assertDictEqual(json.loads(validated_filter_json), default_values)


class StoryModelTest(ModelBaseTest):
    """Tests the user model."""

    def test_story_model(self):
        """
        Test that the test story has the expected data stored in the
        database.
        """
        expected_result = frozenset(
            [('title', 'Test'), ('content', 'Test'), ('user', self.user1),
             ('sketch', self.sketch1)])
        self._test_db_object(expected_result=expected_result, model_cls=Story)



# Copyright 2020 Google Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""This module holds utility functions for the version 1 of the API."""
import logging
import json
import time

from flask import abort
from flask import jsonify

import altair as alt

from timesketch.lib import ontology
from timesketch.lib.aggregators import manager as aggregator_manager
from timesketch.lib.definitions import HTTP_STATUS_CODE_BAD_REQUEST


logger = logging.getLogger('timesketch.api_utils')


def bad_request(message):
    """Function to set custom error message for HTTP 400 requests.

    Args:
        message: Message as string to return to the client.

    Returns: Response object (instance of flask.wrappers.Response)

    """
    response = jsonify({'message': message})
    response.status_code = HTTP_STATUS_CODE_BAD_REQUEST
    return response


def get_sketch_attributes(sketch):
    """Returns a list of attributes of a sketch."""
    attributes = []
    ontology_def = ontology.ONTOLOGY
    for attribute in sketch.attributes:
        if attribute.sketch_id != sketch.id:
            continue
        name = attribute.name
        attribute_values = []
        ontology_string = attribute.ontology
        ontology_dict = ontology_def.get(ontology_string, {})
        cast_as_str = ontology_dict.get('cast_as', 'str')

        for attr_value in attribute.values:
            try:
                value = ontology.cast_variable(attr_value.value, cast_as_str)
            except TypeError:
                value = 'Unable to cast'

            attribute_values.append(value)

        if len(attribute_values) == 1:
            attributes.append(
                (name, attribute_values[0], ontology_string))
        else:
            attributes.append(
                (name, attribute_values, ontology_string))
    return attributes


def run_aggregator(sketch_id, aggregator_name, aggregator_parameters=None,
                   index=None):
    """Run an aggregator and return back results.

    Args:
        sketch_id (int): the sketch ID.
        aggregator_name (str): the name of the aggregator class to run.
        aggregator_parameters (dict): dict containing the parameters used
            for running the aggregator.
        index (list): the list of elasticsearch index names to use.

    Returns:
        Tuple[Object, Dict]: a tuple containing the aggregator result object
            (instance of AggregationResult) and a dict containing metadata
            from the aggregator run.
    """
    agg_class = aggregator_manager.AggregatorManager.get_aggregator(
        aggregator_name)
    if not agg_class:
        return None, {}
    if not aggregator_parameters:
        aggregator_parameters = {}

    aggregator = agg_class(sketch_id=sketch_id, index=index)

    chart_type = aggregator_parameters.pop('supported_charts', None)
    chart_color = aggregator_parameters.pop('chart_color', '')

    time_before = time.time()
    result_obj = aggregator.run(**aggregator_parameters)
    time_after = time.time()

    aggregator_description = aggregator.describe

    meta = {
        'method': 'aggregator_run',
        'chart_type': chart_type,
        'chart_color': chart_color,
        'name': aggregator_description.get('name'),
        'description': aggregator_description.get('description'),
        'es_time': time_after - time_before,
    }

    if chart_type:
        meta['vega_spec'] = result_obj.to_chart(
            chart_name=chart_type,
            chart_title=aggregator.chart_title, color=chart_color)
        meta['vega_chart_title'] = aggregator.chart_title

    return result_obj, meta


def run_aggregator_group(group, sketch_id):
    """Run an aggregator group and return back results.

    Args:
        group (models.sketch.Group): a group object.
        sketch_id (int): the sketch ID.

    Returns:
        Tuple[Object, List, Dict]: a tuple containing the altair chart object,
            a list of result object dicts and a dict containing metadata from
            the aggregator group run.
    """
    result_chart = None
    orientation = group.orientation
    objects = []
    time_before = time.time()
    for aggregator in group.aggregations:
        if aggregator.aggregationgroup_id != group.id:
            abort(
                HTTP_STATUS_CODE_BAD_REQUEST,
                'All aggregations in a group must belong to the group.')
        if aggregator.sketch_id != group.sketch_id:
            abort(
                HTTP_STATUS_CODE_BAD_REQUEST,
                'All aggregations in a group must belong to the group '
                'sketch')

        if aggregator.parameters:
            aggregator_parameters = json.loads(aggregator.parameters)
        else:
            aggregator_parameters = {}

        agg_class = aggregator_manager.AggregatorManager.get_aggregator(
            aggregator.agg_type)
        if not agg_class:
            continue
        aggregator_obj = agg_class(sketch_id=sketch_id)
        chart_type = aggregator_parameters.pop('supported_charts', None)
        color = aggregator_parameters.pop('chart_color', '')
        result_obj = aggregator_obj.run(**aggregator_parameters)

        chart = result_obj.to_chart(
            chart_name=chart_type,
            chart_title=aggregator_obj.chart_title,
            as_chart=True, interactive=True, color=color)

        if result_chart is None:
            result_chart = chart
        elif orientation == 'horizontal':
            result_chart = alt.hconcat(chart, result_chart)
        elif orientation == 'vertical':
            result_chart = alt.vconcat(chart, result_chart)
        else:
            result_chart = alt.layer(chart, result_chart)

        buckets = result_obj.to_dict()
        buckets['buckets'] = buckets.pop('values')
        result = {
            'aggregation_result': {
                aggregator.name: buckets
            }
        }
        objects.append(result)

    parameters = {}
    if group.parameters:
        parameters = json.loads(group.parameters)

    result_chart.title = parameters.get('chart_title', group.name)
    time_after = time.time()

    meta = {
        'method': 'aggregator_group',
        'chart_type': 'compound: {0:s}'.format(orientation),
        'name': group.name,
        'description': group.description,
        'es_time': time_after - time_before,
        'vega_spec': result_chart.to_dict(),
        'vega_chart_title': group.name
    }

    return result_chart, objects, meta
